#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import struct
import time

try:
    import pyarrow as pa
except (AttributeError, ImportError):
    pa = None
try:
    import pyarrow.compute as pac
except (AttributeError, ImportError):
    pac = None
try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None

from ..pb.encoder import Encoder
from ..pb.wire_format import (
    WIRETYPE_VARINT,
    WIRETYPE_FIXED32,
    WIRETYPE_FIXED64,
    WIRETYPE_LENGTH_DELIMITED,
)
from ... import compat, options, types, utils
from ...compat import Enum, futures, six
from ..checksum import Checksum
from ..errors import TunnelError
from ..wireconstants import ProtoWireConstants
from .stream import RequestsIO, get_compress_stream
from .types import odps_schema_to_arrow_schema
try:
    if not options.force_py:
        from ..hasher_c import RecordHasher
        from .writer_c import BaseRecordWriter
    else:
        from ..hasher import RecordHasher
        BaseRecordWriter = None
except ImportError as e:
    if options.force_c:
        raise e
    BaseRecordWriter = RecordHasher = None


varint_tag_types = types.integer_types + (
    types.boolean,
    types.datetime,
    types.date,
    types.interval_year_month,
)
length_delim_tag_types = (
    types.string,
    types.binary,
    types.timestamp,
    types.timestamp_ntz,
    types.interval_day_time,
    types.json,
)


if BaseRecordWriter is None:

    class ProtobufWriter(object):
        """
        ProtobufWriter is a stream-interface wrapper around encoder_c.Encoder(c)
        and encoder.Encoder(py)
        """

        DEFAULT_BUFFER_SIZE = 4096

        def __init__(self, output, buffer_size=None):
            self._encoder = Encoder()
            self._output = output
            self._buffer_size = buffer_size or self.DEFAULT_BUFFER_SIZE
            self._n_total = 0

        def _re_init(self, output):
            self._encoder = Encoder()
            self._output = output
            self._n_total = 0

        def _mode(self):
            return 'py'

        def flush(self):
            if len(self._encoder) > 0:
                data = self._encoder.tostring()
                self._output.write(data)
                self._n_total += len(self._encoder)
                self._encoder = Encoder()

        def close(self):
            self.flush_all()

        def flush_all(self):
            self.flush()
            self._output.flush()

        def _refresh_buffer(self):
            """Control the buffer size of _encoder. Flush if necessary"""
            if len(self._encoder) > self._buffer_size:
                self.flush()

        @property
        def n_bytes(self):
            return self._n_total + len(self._encoder)

        def __len__(self):
            return self.n_bytes

        def _write_tag(self, field_num, wire_type):
            self._encoder.append_tag(field_num, wire_type)
            self._refresh_buffer()

        def _write_raw_long(self, val):
            self._encoder.append_sint64(val)
            self._refresh_buffer()

        def _write_raw_int(self, val):
            self._encoder.append_sint32(val)
            self._refresh_buffer()

        def _write_raw_uint(self, val):
            self._encoder.append_uint32(val)
            self._refresh_buffer()

        def _write_raw_bool(self, val):
            self._encoder.append_bool(val)
            self._refresh_buffer()

        def _write_raw_float(self, val):
            self._encoder.append_float(val)
            self._refresh_buffer()

        def _write_raw_double(self, val):
            self._encoder.append_double(val)
            self._refresh_buffer()

        def _write_raw_string(self, val):
            self._encoder.append_string(val)
            self._refresh_buffer()

    class BaseRecordWriter(ProtobufWriter):
        def __init__(self, schema, out, encoding="utf-8"):
            self._encoding = encoding
            self._schema = schema
            self._columns = self._schema.columns
            self._crc = Checksum()
            self._crccrc = Checksum()
            self._curr_cursor = 0
            self._to_milliseconds = utils.MillisecondsConverter().to_milliseconds
            self._to_milliseconds_utc = utils.MillisecondsConverter(
                local_tz=False
            ).to_milliseconds
            self._to_days = utils.to_days

            super(BaseRecordWriter, self).__init__(out)

        def write(self, record):
            n_record_fields = len(record)
            n_columns = len(self._columns)

            if n_record_fields > n_columns:
                raise IOError("record fields count is more than schema.")

            for i in range(min(n_record_fields, n_columns)):
                if self._schema.is_partition(self._columns[i]):
                    continue

                val = record[i]
                if val is None:
                    continue

                pb_index = i + 1
                self._crc.update_int(pb_index)

                data_type = self._columns[i].type
                if data_type in varint_tag_types:
                    self._write_tag(pb_index, WIRETYPE_VARINT)
                elif data_type == types.float_:
                    self._write_tag(pb_index, WIRETYPE_FIXED32)
                elif data_type == types.double:
                    self._write_tag(pb_index, WIRETYPE_FIXED64)
                elif data_type in length_delim_tag_types:
                    self._write_tag(pb_index, WIRETYPE_LENGTH_DELIMITED)
                elif isinstance(
                    data_type,
                    (
                        types.Char,
                        types.Varchar,
                        types.Decimal,
                        types.Array,
                        types.Map,
                        types.Struct,
                    ),
                ):
                    self._write_tag(pb_index, WIRETYPE_LENGTH_DELIMITED)
                else:
                    raise IOError("Invalid data type: %s" % data_type)
                self._write_field(val, data_type)

            checksum = utils.long_to_int(self._crc.getvalue())
            self._write_tag(ProtoWireConstants.TUNNEL_END_RECORD, WIRETYPE_VARINT)
            self._write_raw_uint(utils.long_to_uint(checksum))
            self._crc.reset()
            self._crccrc.update_int(checksum)
            self._curr_cursor += 1

        def _write_bool(self, data):
            self._crc.update_bool(data)
            self._write_raw_bool(data)

        def _write_long(self, data):
            self._crc.update_long(data)
            self._write_raw_long(data)

        def _write_float(self, data):
            self._crc.update_float(data)
            self._write_raw_float(data)

        def _write_double(self, data):
            self._crc.update_double(data)
            self._write_raw_double(data)

        def _write_string(self, data):
            if isinstance(data, six.text_type):
                data = data.encode(self._encoding)
            self._crc.update(data)
            self._write_raw_string(data)

        def _write_timestamp(self, data, ntz=False):
            to_mills = self._to_milliseconds_utc if ntz else self._to_milliseconds
            t_val = int(to_mills(data.to_pydatetime(warn=False)) / 1000)
            nano_val = data.microsecond * 1000 + data.nanosecond
            self._crc.update_long(t_val)
            self._write_raw_long(t_val)
            self._crc.update_int(nano_val)
            self._write_raw_int(nano_val)

        def _write_interval_day_time(self, data):
            t_val = data.days * 3600 * 24 + data.seconds
            nano_val = data.microseconds * 1000 + data.nanoseconds
            self._crc.update_long(t_val)
            self._write_raw_long(t_val)
            self._crc.update_int(nano_val)
            self._write_raw_int(nano_val)

        def _write_array(self, data, data_type):
            for value in data:
                if value is None:
                    self._write_raw_bool(True)
                else:
                    self._write_raw_bool(False)
                    self._write_field(value, data_type)

        def _write_struct(self, data, data_type):
            if isinstance(data, dict):
                vals = [None] * len(data)
                for idx, key in enumerate(data_type.field_types.keys()):
                    vals[idx] = data[key]
                data = tuple(vals)
            for value, typ in zip(data, data_type.field_types.values()):
                if value is None:
                    self._write_raw_bool(True)
                else:
                    self._write_raw_bool(False)
                    self._write_field(value, typ)

        def _write_field(self, val, data_type):
            if data_type == types.boolean:
                self._write_bool(val)
            elif data_type == types.datetime:
                val = self._to_milliseconds(val)
                self._write_long(val)
            elif data_type == types.date:
                val = self._to_days(val)
                self._write_long(val)
            elif data_type == types.float_:
                self._write_float(val)
            elif data_type == types.double:
                self._write_double(val)
            elif data_type in types.integer_types:
                self._write_long(val)
            elif data_type == types.string:
                self._write_string(val)
            elif data_type == types.binary:
                self._write_string(val)
            elif data_type == types.timestamp or data_type == types.timestamp_ntz:
                self._write_timestamp(val, ntz=data_type == types.timestamp_ntz)
            elif data_type == types.interval_day_time:
                self._write_interval_day_time(val)
            elif data_type == types.interval_year_month:
                self._write_long(val.total_months())
            elif isinstance(data_type, (types.Char, types.Varchar)):
                self._write_string(val)
            elif isinstance(data_type, types.Decimal):
                self._write_string(str(val))
            elif isinstance(data_type, types.Json):
                self._write_string(json.dumps(val))
            elif isinstance(data_type, types.Array):
                self._write_raw_uint(len(val))
                self._write_array(val, data_type.value_type)
            elif isinstance(data_type, types.Map):
                self._write_raw_uint(len(val))
                self._write_array(compat.lkeys(val), data_type.key_type)
                self._write_raw_uint(len(val))
                self._write_array(compat.lvalues(val), data_type.value_type)
            elif isinstance(data_type, types.Struct):
                self._write_struct(val, data_type)
            else:
                raise IOError("Invalid data type: %s" % data_type)

        @property
        def count(self):
            return self._curr_cursor

        def _write_finish_tags(self):
            self._write_tag(ProtoWireConstants.TUNNEL_META_COUNT, WIRETYPE_VARINT)
            self._write_raw_long(self.count)
            self._write_tag(ProtoWireConstants.TUNNEL_META_CHECKSUM, WIRETYPE_VARINT)
            self._write_raw_uint(utils.long_to_uint(self._crccrc.getvalue()))

        def close(self):
            self._write_finish_tags()
            super(BaseRecordWriter, self).close()
            self._curr_cursor = 0

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            # if an error occurs inside the with block, we do not commit
            if exc_val is not None:
                return
            self.close()


class RecordWriter(BaseRecordWriter):
    """
    This writer uploads the output of serializer asynchronously within a long-lived http connection.
    """

    def __init__(self, schema, request_callback, compress_option=None, encoding="utf-8"):
        self._req_io = RequestsIO(request_callback, chunk_size=options.chunk_size)

        out = get_compress_stream(self._req_io, compress_option)
        super(RecordWriter, self).__init__(schema, out, encoding=encoding)
        self._req_io.start()

    def write(self, record):
        if self._req_io._async_err:
            ex_type, ex_value, tb = self._req_io._async_err
            six.reraise(ex_type, ex_value, tb)
        super(RecordWriter, self).write(record)

    def close(self):
        super(RecordWriter, self).close()
        self._req_io.finish()

    def get_total_bytes(self):
        return self.n_bytes


class BufferedRecordWriter(BaseRecordWriter):
    """
    This writer buffers the output of serializer. When the buffer exceeds a fixed-size of limit
     (default 20 MiB), it uploads the buffered output within one http connection.
    """
    def __init__(
        self,
        schema,
        request_callback,
        compress_option=None,
        encoding="utf-8",
        buffer_size=None,
        block_id=None,
        block_id_gen=None,
    ):
        self._buffer_size = buffer_size or options.tunnel.block_buffer_size
        self._request_callback = request_callback
        self._block_id = block_id or 0
        self._blocks_written = []
        self._buffer = compat.BytesIO()
        self._n_bytes_written = 0
        self._compress_option = compress_option
        self._block_id_gen = block_id_gen

        out = get_compress_stream(self._buffer, compress_option)
        super(BufferedRecordWriter, self).__init__(schema, out, encoding=encoding)

    @property
    def cur_block_id(self):
        return self._block_id

    def _get_next_block_id(self):
        if callable(self._block_id_gen):
            return self._block_id_gen()
        return self._block_id + 1

    def write(self, record):
        super(BufferedRecordWriter, self).write(record)
        if 0 < self._buffer_size < self._n_raw_bytes:
            self._flush()

    def close(self):
        if self._n_raw_bytes > 0:
            self._flush()
        self.flush_all()
        self._buffer.close()

    def _reset_writer(self, write_response):
        self._buffer = compat.BytesIO()

        out = get_compress_stream(self._buffer, self._compress_option)
        self._re_init(out)
        self._curr_cursor = 0
        self._crccrc.reset()
        self._crc.reset()

    def _send_buffer(self):
        def gen():  # synchronize chunk upload
            data = self._buffer.getvalue()
            chunk_size = options.chunk_size
            while data:
                to_send = data[:chunk_size]
                data = data[chunk_size:]
                yield to_send

        return self._request_callback(self._block_id, gen())

    def _flush(self):
        self._write_finish_tags()
        self._n_bytes_written += self._n_raw_bytes
        self.flush_all()

        resp = self._send_buffer()
        self._blocks_written.append(self._block_id)
        self._block_id = self._get_next_block_id()

        self._reset_writer(resp)

    @property
    def _n_raw_bytes(self):
        return super(BufferedRecordWriter, self).n_bytes

    @property
    def n_bytes(self):
        return self._n_bytes_written + self._n_raw_bytes

    def get_total_bytes(self):
        return self.n_bytes

    def get_blocks_written(self):
        return self._blocks_written


# make sure original typo class also referable
BufferredRecordWriter = BufferedRecordWriter


class StreamRecordWriter(BufferedRecordWriter):
    def __init__(
        self,
        schema,
        request_callback,
        session,
        slot,
        compress_option=None,
        encoding="utf-8",
        buffer_size=None,
    ):
        self.session = session
        self.slot = slot
        self._record_count = 0

        super(StreamRecordWriter, self).__init__(
            schema,
            request_callback,
            compress_option=compress_option,
            encoding=encoding,
            buffer_size=buffer_size,
        )

    def write(self, record):
        super(StreamRecordWriter, self).write(record)
        self._record_count += 1

    def _reset_writer(self, write_response):
        self._record_count = 0
        slot_server = write_response.headers['odps-tunnel-routed-server']
        slot_num = int(write_response.headers['odps-tunnel-slot-num'])
        self.session.reload_slots(self.slot, slot_server, slot_num)
        super(StreamRecordWriter, self)._reset_writer(write_response)

    def _send_buffer(self):
        def gen():  # synchronize chunk upload
            data = self._buffer.getvalue()
            chunk_size = options.chunk_size
            while data:
                to_send = data[:chunk_size]
                data = data[chunk_size:]
                yield to_send

        return self._request_callback(gen())


class BaseArrowWriter(object):
    def __init__(self, schema, out=None, chunk_size=None):
        if pa is None:
            raise ValueError("To use arrow writer you need to install pyarrow")

        self._schema = schema
        self._arrow_schema = odps_schema_to_arrow_schema(schema)
        self._chunk_size = chunk_size or options.chunk_size
        self._crc = Checksum()
        self._crccrc = Checksum()
        self._cur_chunk_size = 0

        self._output = out
        self._write_chunk_size()

    def _re_init(self, output):
        self._output = output

    def _write_chunk_size(self):
        self._write_unint32(self._chunk_size)

    def _write_unint32(self, val):
        data = struct.pack("!I", utils.long_to_uint(val))
        self._output.write(data)

    def _write_chunk(self, buf):
        self._output.write(buf)
        self._crccrc.update(buf)
        self._cur_chunk_size += len(buf)
        if self._cur_chunk_size >= self._chunk_size:
            self._crc.update(buf)
            checksum = self._crc.getvalue()
            self._write_unint32(checksum)
            self._crc.reset()
            self._cur_chunk_size = 0

    @classmethod
    def _localize_timezone(cls, col, tz=None):
        from ...lib import tzlocal

        if tz is None:
            if options.local_timezone is True or options.local_timezone is None:
                tz = str(tzlocal.get_localzone())
            elif options.local_timezone is False:
                tz = "UTC"
            else:
                tz = str(options.local_timezone)

        try:
            if col.type.tz is not None:
                return col
            col = pac.assume_timezone(col, tz)
            return col
        except:
            col = col.to_pandas()
        col = pa.Array.from_pandas(col.dt.tz_localize(tz))
        return col

    def write(self, data):
        if isinstance(data, pd.DataFrame):
            arrow_data = pa.RecordBatch.from_pandas(data)
        elif isinstance(data, (pa.Table, pa.RecordBatch)):
            arrow_data = data
        else:
            raise TypeError("Cannot support writing data type %s", type(data))

        assert isinstance(arrow_data, (pa.RecordBatch, pa.Table))

        if not arrow_data.schema.equals(self._arrow_schema):
            type_dict = dict(zip(arrow_data.schema.names, arrow_data.schema.types))
            column_dict = dict(zip(arrow_data.schema.names, arrow_data.columns))
            arrays = []
            for name, tp in zip(self._arrow_schema.names, self._arrow_schema.types):
                if name not in column_dict:
                    raise ValueError("Input record batch does not contain column %s" % name)

                if tp == pa.timestamp("ms") or tp == pa.timestamp("ns"):
                    if self._schema[name].type == types.timestamp_ntz:
                        column_dict[name] = self._localize_timezone(column_dict, "UTC")
                    else:
                        column_dict[name] = self._localize_timezone(column_dict[name])

                if tp == type_dict[name]:
                    arrays.append(column_dict[name])
                else:
                    try:
                        arrays.append(column_dict[name].cast(tp, safe=False))
                    except (pa.ArrowInvalid, pa.ArrowNotImplementedError):
                        raise ValueError("Failed to cast column %s to type %s" % (name, tp))
            pa_type = type(arrow_data)
            arrow_data = pa_type.from_arrays(arrays, names=self._arrow_schema.names)

        if isinstance(arrow_data, pa.RecordBatch):
            batches = [arrow_data]
        else:  # pa.Table
            batches = arrow_data.to_batches()

        for batch in batches:
            data = batch.serialize().to_pybytes()
            written_bytes = 0
            while written_bytes < len(data):
                length = min(self._chunk_size - self._cur_chunk_size,
                             len(data) - written_bytes)
                chunk_data = data[written_bytes: written_bytes + length]
                self._write_chunk(chunk_data)
                written_bytes += length

    def _write_finish_tags(self):
        checksum = self._crccrc.getvalue()
        self._write_unint32(checksum)
        self._crccrc.reset()

    def flush(self):
        self._output.flush()

    def _finish(self):
        self._write_finish_tags()
        self._output.flush()

    def close(self):
        self._finish()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # if an error occurs inside the with block, we do not commit
        if exc_val is not None:
            return
        self.close()


class ArrowWriter(BaseArrowWriter):
    def __init__(
        self, schema, request_callback, compress_option=None, chunk_size=None
    ):
        self._req_io = RequestsIO(request_callback, chunk_size=chunk_size)

        out = get_compress_stream(self._req_io, compress_option)
        super(ArrowWriter, self).__init__(schema, out, chunk_size)
        self._req_io.start()

    def _finish(self):
        super(ArrowWriter, self)._finish()
        self._req_io.finish()


class BufferedArrowWriter(BaseArrowWriter):
    def __init__(
        self,
        schema,
        request_callback,
        compress_option=None,
        buffer_size=None,
        chunk_size=None,
        block_id=None,
        block_id_gen=None,
    ):
        self._buffer_size = buffer_size or options.tunnel.block_buffer_size
        self._request_callback = request_callback
        self._block_id = block_id or 0
        self._blocks_written = []
        self._buffer = compat.BytesIO()
        self._compress_option = compress_option
        self._n_bytes_written = 0
        self._block_id_gen = block_id_gen

        out = get_compress_stream(self._buffer, compress_option)
        super(BufferedArrowWriter, self).__init__(schema, out, chunk_size=chunk_size)

    @property
    def cur_block_id(self):
        return self._block_id

    def _get_next_block_id(self):
        if callable(self._block_id_gen):
            return self._block_id_gen()
        return self._block_id + 1

    def write(self, data):
        super(BufferedArrowWriter, self).write(data)
        if 0 < self._buffer_size < self._n_raw_bytes:
            self._flush()

    def close(self):
        if self._n_raw_bytes > 0:
            self._flush()
        self._finish()
        self._buffer.close()

    def _reset_writer(self):
        self._buffer = compat.BytesIO()

        out = get_compress_stream(self._buffer, self._compress_option)
        self._re_init(out)
        self._crccrc.reset()
        self._crc.reset()

    def _send_buffer(self):
        def gen():  # synchronize chunk upload
            data = self._buffer.getvalue()
            chunk_size = options.chunk_size
            while data:
                to_send = data[:chunk_size]
                data = data[chunk_size:]
                yield to_send

        return self._request_callback(self._block_id, gen())

    def _flush(self):
        self._write_finish_tags()
        self._n_bytes_written += self._n_raw_bytes

        self._send_buffer()
        self._blocks_written.append(self._block_id)
        self._block_id = self._get_next_block_id()

        self._reset_writer()

    @property
    def _n_raw_bytes(self):
        return self._buffer.tell()

    @property
    def n_bytes(self):
        return self._n_bytes_written + self._n_raw_bytes

    def get_total_bytes(self):
        return self.n_bytes

    def get_blocks_written(self):
        return self._blocks_written


class Upsert(object):
    DEFAULT_MAX_BUFFER_SIZE = 64 * 1024 ** 2
    DEFAULT_SLOT_BUFFER_SIZE = 1024 ** 2

    class Operation(Enum):
        UPSERT = "UPSERT"
        DELETE = "DELETE"

    class Status(Enum):
        NORMAL = "NORMAL"
        ERROR = "ERROR"
        CLOSED = "CLOSED"

    def __init__(
        self,
        schema,
        request_callback,
        session,
        compress_option=None,
        encoding="utf-8",
        max_buffer_size=None,
        slot_buffer_size=None,
    ):
        self._schema = schema
        self._request_callback = request_callback
        self._session = session
        self._compress_option = compress_option

        self._max_buffer_size = max_buffer_size or self.DEFAULT_MAX_BUFFER_SIZE
        self._slot_buffer_size = slot_buffer_size or self.DEFAULT_SLOT_BUFFER_SIZE
        self._total_n_bytes = 0

        self._status = Upsert.Status.NORMAL

        self._schema = session.schema
        self._encoding = encoding
        self._hash_keys = self._session.hash_keys
        self._hasher = RecordHasher(schema, self._session.hasher, self._hash_keys)
        self._buckets = self._session.buckets.copy()
        self._bucket_buffers = {}
        self._bucket_writers = {}

        for slot in session.buckets.keys():
            self._build_bucket_writer(slot)

    @property
    def status(self):
        return self._status

    @property
    def n_bytes(self):
        return self._total_n_bytes

    def upsert(self, record):
        return self._write(record, Upsert.Operation.UPSERT)

    def delete(self, record):
        return self._write(record, Upsert.Operation.DELETE)

    def flush(self, flush_all=True):
        if len(self._session.buckets) != len(self._bucket_writers):
            raise TunnelError("session slot map is changed")
        else:
            self._buckets = self._session.buckets.copy()

        bucket_written = dict()
        bucket_to_count = dict()

        def write_bucket(bucket_id):
            slot = self._buckets[bucket_id]
            sio = self._bucket_buffers[bucket_id]
            rec_count = bucket_to_count[bucket_id]

            self._request_callback(bucket_id, slot, rec_count, sio.getvalue())
            self._build_bucket_writer(bucket_id)
            bucket_written[bucket_id] = True

        retry = 0
        while True:
            futs = []
            pool = futures.ThreadPoolExecutor(len(self._bucket_writers))
            try:
                self._check_status()
                for bucket, writer in self._bucket_writers.items():
                    if writer.n_bytes == 0 or bucket_written.get(bucket):
                        continue
                    if not flush_all and writer.n_bytes <= self._slot_buffer_size:
                        continue

                    bucket_to_count[bucket] = writer.count
                    writer.close()
                    futs.append(pool.submit(write_bucket, bucket))
                for fut in futs:
                    fut.result()
                break
            except KeyboardInterrupt:
                raise TunnelError("flush interrupted")
            except:
                retry += 1
                if retry == 3:
                    raise
            finally:
                pool.shutdown()

    def close(self):
        if self.status == Upsert.Status.NORMAL:
            self.flush()
            self._status = Upsert.Status.CLOSED

    def _build_bucket_writer(self, slot):
        self._bucket_buffers[slot] = compat.BytesIO()
        self._bucket_writers[slot] = BaseRecordWriter(
            self._schema,
            get_compress_stream(self._bucket_buffers[slot], self._compress_option),
            encoding=self._encoding,
        )

    def _check_status(self):
        if self._status == Upsert.Status.CLOSED:
            raise TunnelError("Stream is closed!")
        elif self._status == Upsert.Status.ERROR:
            raise TunnelError("Stream has error!")

    def _write(self, record, op, valid_columns=None):
        self._check_status()

        bucket = self._hasher.hash(record) % len(self._bucket_writers)
        if bucket not in self._bucket_writers:
            raise TunnelError(
                "Tunnel internal error! Do not have bucket for hash key " + bucket
            )
        record[self._session.UPSERT_OPERATION_KEY] = ord(
            b"U" if op == Upsert.Operation.UPSERT else b"D"
        )
        if valid_columns is None:
            record[self._session.UPSERT_VALUE_COLS_KEY] = []
        else:
            valid_cols_set = set(valid_columns)
            col_idxes = [idx for idx, col in self._schema.columns if col in valid_cols_set]
            record[self._session.UPSERT_VALUE_COLS_KEY] = col_idxes

        writer = self._bucket_writers[bucket]
        prev_written_size = writer.n_bytes
        writer.write(record)
        written_size = writer.n_bytes
        self._total_n_bytes += written_size - prev_written_size

        if writer.n_bytes > self._slot_buffer_size:
            self.flush(False)
        elif self._total_n_bytes > self._max_buffer_size:
            self.flush(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
import operator

from ..rewriter import BaseRewriter
from ...expr.merge import *
from ...utils import traverse_until_source


class Rewriter(BaseRewriter):
    def visit_project_collection(self, expr):
        self._rewrite_reduction_in_projection(expr)

    def visit_filter_collection(self, expr):
        self._rewrite_reduction_in_filter(expr)

    def visit_column(self, expr):
        if isinstance(expr.input, JoinCollectionExpr):
            input = expr.input
            name = expr.source_name
            while isinstance(input, JoinCollectionExpr):
                idx, name = input._column_origins[name]
                input = (input.lhs, input.rhs)[idx]

            self._sub(expr, input[name].rename(expr.name))

    def visit_join(self, expr):
        if expr._predicate and isinstance(expr._predicate, list):
            expr._predicate = reduce(operator.and_, expr._predicate)

        for node in (expr.rhs,):
            parents = self._parents(node)
            if isinstance(node, JoinCollectionExpr):
                projection = JoinProjectCollectionExpr(
                    _input=node, _schema=node.schema,
                    _fields=node._fetch_fields())
                self._sub(node, projection, parents)
            elif isinstance(node, JoinProjectCollectionExpr):
                self._sub(node.input, node, parents)

        need_project = [False, ]

        def walk(node):
            if isinstance(node, JoinCollectionExpr) and \
                    node.column_conflict:
                need_project[0] = True
                return

            if isinstance(node, JoinCollectionExpr):
                walk(node.lhs)

        walk(expr)

        if need_project[0]:
            parents = self._parents(expr)
            if not parents or \
                    not any(isinstance(parent, (ProjectCollectionExpr, JoinCollectionExpr))
                            for parent in parents):
                to_sub = expr[expr]
                self._sub(expr, to_sub, parents)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

try:
    import pandas as pd
except ImportError:
    pass

from .core import Engine
from .odpssql.engine import ODPSSQLEngine
from .pd.engine import PandasEngine
from .seahawks.engine import SeahawksEngine
from .seahawks.models import SeahawksTable
from .sqlalchemy.engine import SQLAlchemyEngine
from .selecter import available_engines, Engines, EngineSelecter
from .formatter import ExprExecutionGraphFormatter
from .context import context
from .utils import process_persist_kwargs
from .. import Scalar
from ..expr.core import ExprDAG
from ..expr.expressions import CollectionExpr
from ..expr.merge import JoinCollectionExpr, UnionCollectionExpr
from ..expr.element import IsIn
from ...models import Table
from ... import options
from ...utils import gen_repr_object


def get_default_engine(*exprs):
    from ... import ODPS

    odps = None
    engines = set()

    for expr in exprs:
        if expr._engine:
            return expr._engine

        srcs = list(expr.data_source())
        expr_engines = list(available_engines(srcs))
        engines.update(set(expr_engines))

        if len(expr_engines) == 1:
            engine = expr_engines[0]
            src = srcs[0]
            if engine in (Engines.ODPS, Engines.ALGO):
                expr_odps = src.project.odps
            elif engine in (Engines.PANDAS, Engines.SQLALCHEMY):
                expr_odps = None
            else:
                raise NotImplementedError
        else:
            table_src = next(it for it in srcs if hasattr(it, 'project'))
            expr_odps = table_src.project.odps
        if expr_odps is not None:
            odps = expr_odps

    if odps is None and options.account is not None and \
            options.endpoint is not None and options.default_project is not None:
        odps = ODPS._from_account(
            options.account, options.default_project,
            endpoint=options.endpoint,
            tunnel_endpoint=options.tunnel.endpoint,
            overwrite_global=False,
        )

    return MixedEngine(odps, list(engines))


class MixedEngine(Engine):
    def __init__(self, odps, engines=None):
        self._odps = odps
        self._engines = engines
        self._generated_table_names = []

        self._selecter = EngineSelecter()

        self._pandas_engine = PandasEngine(self._odps)
        self._odpssql_engine = ODPSSQLEngine(self._odps)
        self._seahawks_engine = SeahawksEngine(self._odps)
        self._sqlalchemy_engine = SQLAlchemyEngine(self._odps)

        from ...ml.engine import OdpsAlgoEngine
        self._xflow_engine = OdpsAlgoEngine(self._odps)

    def stop(self):
        self._pandas_engine.stop()
        self._odpssql_engine.stop()
        self._seahawks_engine.stop()
        self._xflow_engine.stop()

    def _gen_table_name(self):
        table_name = self._odpssql_engine._gen_table_name()
        self._generated_table_names.append(table_name)
        return table_name

    def _get_backend(self, expr_dag):
        engine = self._selecter.select(expr_dag)
        if engine == Engines.ODPS:
            return self._odpssql_engine
        elif engine == Engines.PANDAS:
            return self._pandas_engine
        elif engine == Engines.SQLALCHEMY:
            return self._sqlalchemy_engine
        elif engine == Engines.ALGO:
            return self._xflow_engine
        else:
            assert engine == Engines.SEAHAWKS
            return self._seahawks_engine

    def _delegate(self, method, expr_dag, dag, expr, **kwargs):
        return getattr(self._get_backend(expr_dag), method)(expr_dag, dag, expr, **kwargs)

    def _cache(self, expr_dag, dag, expr, **kwargs):
        return self._delegate('_cache', expr_dag, dag, expr, **kwargs)

    def _handle_dep(self, expr_dag, dag, expr, **kwargs):
        return self._delegate('_handle_dep', expr_dag, dag, expr, **kwargs)

    def _execute(self, expr_dag, dag, expr, **kwargs):
        return self._delegate('_execute', expr_dag, dag, expr, **kwargs)

    def _persist(self, name, expr_dag, dag, expr, **kwargs):
        return self._get_backend(expr_dag)._persist(name, expr_dag, dag, expr, **kwargs)

    def _handle_join_or_union(self, expr_dag, dag, _, **kwargs):
        root = expr_dag.root
        if not self._selecter.has_diff_data_sources(root, no_cache=True):
            return

        to_execute = root._lhs if not self._selecter.has_odps_data_source(root._lhs) \
            else root._rhs
        table_name = self._gen_table_name()
        odps_table = self._odps.get_table(table_name, schema=self._odpssql_engine._ctx.default_schema)
        sub = CollectionExpr(_source_data=odps_table, _schema=to_execute.schema)
        sub.add_deps(to_execute)
        expr_dag.substitute(to_execute, sub)

        # prevent the kwargs come from `persist`
        process_persist_kwargs(kwargs)

        execute_dag = ExprDAG(to_execute, dag=expr_dag)
        return self._get_backend(execute_dag)._persist(
            table_name, execute_dag, dag, to_execute, **kwargs)

    def _handle_isin(self, expr_dag, dag, expr, **kwargs):
        if not self._selecter.has_diff_data_sources(expr_dag.root, no_cache=True):
            return

        seq = expr._values[0]
        expr._values = None
        execute_dag = ExprDAG(seq, dag=expr_dag)
        execute_node = self._get_backend(execute_dag)._execute(
            execute_dag, dag, seq, **kwargs)

        def callback(res):
            vals = res[:, 0].tolist()
            expr._values = tuple(Scalar(val) for val in vals)
        execute_node.callback = callback

        return execute_node

    def _handle_function(self, expr_dag, dag, _, **kwargs):
        root = expr_dag.root

        # if expr input comes from an ODPS table
        is_root_input_from_odps = \
            self._selecter.has_odps_data_source(root.children()[0])

        for i, collection in enumerate(root._collection_resources):
            # if collection resource comes from an ODPS table
            is_source_from_odps = self._selecter.has_odps_data_source(collection)

            if is_root_input_from_odps and not is_source_from_odps:
                table_name = self._gen_table_name()
                odps_table = self._odps.get_table(table_name, schema=self._odpssql_engine._ctx.default_schema)
                sub = CollectionExpr(_source_data=odps_table, _schema=collection.schema)
                sub.add_deps(collection)
                expr_dag.substitute(collection, sub)

                # prevent the kwargs come from `persist`
                process_persist_kwargs(kwargs)

                execute_dag = ExprDAG(collection, dag=expr_dag)
                self._get_backend(execute_dag)._persist(
                    table_name, execute_dag, dag, collection, **kwargs)
            elif not is_root_input_from_odps and is_source_from_odps:
                if not self._selecter.has_pandas_data_source(root.children()[0]):
                    raise NotImplementedError
                sub = CollectionExpr(_source_data=pd.DataFrame(),
                                     _schema=collection.schema)
                sub.add_deps(collection)
                expr_dag.substitute(collection, sub)
                execute_node = self._odpssql_engine._execute(
                    ExprDAG(collection, dag=expr_dag), dag, collection, **kwargs)

                def callback(res):
                    sub._source_data = res.values
                execute_node.callback = callback
            else:
                continue

    def _dispatch(self, expr_dag, expr, ctx):
        from ...ml.expr import AlgoExprMixin
        funcs = []

        if isinstance(expr, AlgoExprMixin):
            return self._xflow_engine._dispatch(expr_dag, expr, ctx)

        handle = None
        if isinstance(expr, (JoinCollectionExpr, UnionCollectionExpr)) and \
                self._selecter.has_diff_data_sources(expr):
            handle = self._handle_join_or_union
        elif isinstance(expr, IsIn) and \
                self._selecter.has_diff_data_sources(expr):
            handle = self._handle_isin
        elif hasattr(expr, '_func') and \
                getattr(expr, '_collection_resources', None) is not None and \
                self._selecter.has_diff_data_sources(expr):
            handle = self._handle_function
        if handle is not None:
            funcs.append(handle)

        h = super(MixedEngine, self)._dispatch(expr_dag, expr, ctx)
        if h is not None:
            if handle is None:
                return h
            funcs.append(h)

        if funcs:
            def f(*args, **kwargs):
                for func in funcs:
                    func(*args, **kwargs)

            return f

    def _get_cached_sub_expr(self, cached_expr, ctx=None):
        ctx = ctx or context
        data = ctx.get_cached(cached_expr)
        if self._selecter.force_odps and isinstance(data, SeahawksTable):
            # skip seahawks heap table
            return
        return super(MixedEngine, self)._get_cached_sub_expr(cached_expr, ctx=ctx)

    def visualize(self, expr):
        dag = self.compile(expr)
        try:
            formatter = ExprExecutionGraphFormatter(dag)
            return gen_repr_object(svg=formatter._repr_svg_())
        finally:
            self._generated_table_names = []

    def compile(self, *expr_args_kwargs):
        src_expr_args_kwargs = tuple(expr_args_kwargs)
        exprs_dags, expr_args_kwargs = self._process(*expr_args_kwargs)

        if not self._selecter.force_odps and any(it[0] == '_persist' for it in expr_args_kwargs):
            self._selecter.force_odps = True
            return self.compile(*src_expr_args_kwargs)

        engine_types = [self._selecter.select(expr_dag) for expr_dag in exprs_dags]
        if any(engine_type == Engines.ODPS for engine_type in engine_types):
            self._selecter.force_odps = True
            if any(engine_type == Engines.SEAHAWKS for engine_type in engine_types):
                return self.compile(*src_expr_args_kwargs)

        try:
            dag = super(MixedEngine, self)._compile_dag(expr_args_kwargs, exprs_dags)
        except NotImplementedError:
            self._selecter.force_odps = True
            return self.compile(*src_expr_args_kwargs)

        if any(engine_type == Engines.SEAHAWKS for engine_type in engine_types):
            def fallback():
                self._selecter.force_odps = True
                return self.compile(*src_expr_args_kwargs)

            def need_fallback(e):
                try:
                    import sqlalchemy
                    exceptions = (NotImplementedError, sqlalchemy.exc.DatabaseError)
                except ImportError:
                    exceptions = (NotImplementedError,)

                if not isinstance(e, exceptions):
                    return False
                if not isinstance(e, NotImplementedError) and \
                        'AXF Exception' not in str(e):
                    # the seahawks error
                    return False
                return True

            if not all(not hasattr(n, 'verify') or n.verify()
                       for n in dag.indep_nodes()):
                # we only verify the independent nodes
                dag = fallback()

            dag.fallback = fallback
            dag.need_fallback = need_fallback
        return dag



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import calendar
import struct

from .. import types
from ..compat import PY27
from ..utils import to_binary, to_milliseconds

_int32_struct = struct.Struct("<l")
_int64_struct = struct.Struct("<q")
_float_struct = struct.Struct("<f")
_double_struct = struct.Struct("<d")

_ord_code = ord if PY27 else (lambda x: x)


class AbstractHasher(object):
    def hash_bigint(self, val):
        raise NotImplementedError

    def hash_float(self, val):
        raise NotImplementedError

    def hash_double(self, val):
        raise NotImplementedError

    def hash_bool(self, val):
        raise NotImplementedError

    def hash_string(self, val):
        raise NotImplementedError


class DefaultHasher(AbstractHasher):
    def hash_bigint(self, val):
        val = (~val) + ctypes.c_int64(val << 18).value
        val ^= (val >> 31)
        val = ctypes.c_int64(val * 21).value
        val ^= (val >> 11)
        val += ctypes.c_int64(val << 6).value
        val ^= (val >> 22)
        return ctypes.c_int32(val).value

    def hash_float(self, val):
        return self.hash_bigint(_int32_struct.unpack(_float_struct.pack(val))[0])

    def hash_double(self, val):
        return self.hash_bigint(_int64_struct.unpack(_double_struct.pack(val))[0])

    def hash_bool(self, val):
        # it is a magic number
        if val:
            return 0x172ba9c7
        else:
            return -0x3a59cb12

    def hash_string(self, val):
        val = to_binary(val)
        hash_val = 0
        for ch in val:
            hash_val += _ord_code(ch)
            hash_val += ctypes.c_int32(hash_val << 10).value
            hash_val ^= hash_val >> 6
        hash_val += ctypes.c_int32(hash_val << 3).value
        hash_val ^= hash_val >> 11
        hash_val += ctypes.c_int32(hash_val << 15).value
        return ctypes.c_int32(hash_val).value


class LegacyHasher(AbstractHasher):
    def hash_bigint(self, val):
        return ctypes.c_int32((val >> 32) ^ val).value

    def hash_float(self, val):
        return self.hash_bigint(_int32_struct.unpack(_float_struct.pack(val))[0])

    def hash_double(self, val):
        return self.hash_bigint(_int64_struct.unpack(_double_struct.pack(val))[0])

    def hash_bool(self, val):
        # it is a magic number
        if val:
            return 0x172ba9c7
        else:
            return -0x3a59cb12

    def hash_string(self, val):
        val = to_binary(val)
        hash_val = 0
        for ch in val:
            hash_val = ctypes.c_int32(hash_val * 31 + _ord_code(ch)).value
        return hash_val


def get_hasher(hasher_type):
    if hasher_type == "legacy":
        return LegacyHasher()
    elif hasher_type == "default":
        return DefaultHasher()
    else:
        raise ValueError("Hasher type %s not supported" % hasher_type)


def _hash_date(hasher, x):
    return hasher.hash_bigint(int(calendar.timegm(x.timetuple())))


def _hash_datetime(hasher, x):
    return hasher.hash_bigint(int(to_milliseconds(x)))


def _hash_timestamp(hasher, x):
    seconds = int(to_milliseconds(x.to_pydatetime()) / 1000)
    nanos = x.microsecond * 1000 + x.nanosecond
    return hasher.hash_bigint((seconds << 30) | nanos)


def _hash_timedelta(hasher, x):
    seconds = int(x.total_seconds())
    nanos = x.microseconds * 1000 + x.nanoseconds
    return hasher.hash_bigint((seconds << 30) | nanos)


_type_to_hash_fun = {
    types.tinyint: lambda hasher, x: hasher.hash_bigint(x),
    types.smallint: lambda hasher, x: hasher.hash_bigint(x),
    types.int_: lambda hasher, x: hasher.hash_bigint(x),
    types.bigint: lambda hasher, x: hasher.hash_bigint(x),
    types.boolean: lambda hasher, x: hasher.hash_bool(x),
    types.float_: lambda hasher, x: hasher.hash_float(x),
    types.double: lambda hasher, x: hasher.hash_double(x),
    types.date: _hash_date,
    types.datetime: _hash_datetime,
    types.timestamp: _hash_timestamp,
    types.interval_day_time: _hash_timedelta,
    types.binary: lambda hasher, x: hasher.hash_string(x),
    types.string: lambda hasher, x: hasher.hash_string(x),
}


class RecordHasher(object):
    def __init__(self, schema, hasher_type, hash_keys):
        self._schema = schema
        self._hasher = get_hasher(hasher_type)
        self._hash_keys = hash_keys

        self._column_hash_appenders = []
        for col_name in hash_keys:
            col = self._schema.get_column(col_name)
            if col.type in _type_to_hash_fun:
                self._column_hash_appenders.append(
                    _type_to_hash_fun[col.type]
                )
            elif isinstance(col.type, (types.Char, types.Varchar)):
                self._column_hash_appenders.append(
                    _type_to_hash_fun[types.string]
                )
            else:
                raise TypeError("Hash for type %s not supported" % col.type)

    def hash(self, record):
        hash_sum = 0
        for idx, key in enumerate(self._hash_keys):
            if record[key] is None:
                continue
            hash_sum += self._column_hash_appenders[idx](self._hasher, record[key])
        hash_sum = ctypes.c_int32(hash_sum).value
        return hash_sum ^ (hash_sum >> 8)


def hash_value(hasher_type, data_type, value):
    """Simple value hash function for test purpose"""
    hasher = get_hasher(hasher_type)
    data_type = types.validate_data_type(data_type)
    return _type_to_hash_fun[data_type](hasher, value)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import warnings

from odps.compat import enum

from ..utils import get_environ
from ..rpc import CupidRpcController, SandboxRpcChannel

try:
    from ..proto import cupid_process_service_pb2 as process_pb
except TypeError:
    warnings.warn('Cannot import protos from pycupid: '
        'consider upgrading your protobuf python package.', ImportWarning)
    raise ImportError

logger = logging.getLogger(__name__)
_pid_to_context = dict()


def context():
    try:
        from .ctypes_libs import Subprocess_Container_Init, Subprocess_StartFDReceiver
    except ImportError:
        return

    pid = os.getpid()
    if pid not in _pid_to_context:
        Subprocess_Container_Init()
        Subprocess_StartFDReceiver()
        _pid_to_context[pid] = RuntimeContext()
    return _pid_to_context[pid]


class ContainerStatus(enum.Enum):
    START = 1
    TERMINATED = 2
    RUNNING = 3


class WorkItemProgress(enum.Enum):
    WIP_WAITING = 1      # waiting for enough resource
    WIP_READY = 2        # resource ready, but not running yet
    WIP_RUNNING = 3
    WIP_TERMINATING = 4  # worker has tried to terminate the work item but is not yet confirmed
    WIP_TERMINATED = 5   # work confirms that the work item is terminated normally
    WIP_FAILED = 6       # worker confirms that the work item has failed
    WIP_INTERRUPTED = 7
    WIP_DEAD = 8
    WIP_INTERRUPTING = 9


class RuntimeContext(object):
    def __init__(self):
        from .ctypes_libs import ChannelConf, ChannelSlaveClient
        # init channels
        server_read_pipe_num = int(get_environ('serverReadPipeNum', 2))
        server_write_pipe_num = int(get_environ('serverWritePipeNum', 2))
        client_read_pipe_num = int(get_environ('clientReadPipeNum', 2))
        client_write_pipe_num = int(get_environ('clientWritePipeNum', 2))
        cpu_cores = min(1, int(get_environ('workerCores', 2)))
        channel_conf = ChannelConf()
        channel_conf.set_integer('odps.subprocess.channel.rpc.handler.count',
                                 max(cpu_cores / 2, 1))
        channel_conf.set_integer('odps.subprocess.channel.stream.handler.count', cpu_cores * 3)

        logger.info("serverReadPipeNum: %d, serverWritePipeNum: %d, clientReadPipeNum: %d, "
                    "clientWritePipeNum: %d, cpuCores: %d" %
                    (server_read_pipe_num, server_write_pipe_num, client_read_pipe_num,
                     client_write_pipe_num, cpu_cores))
        self._channel_client = ChannelSlaveClient(client_write_pipe_num, client_read_pipe_num, 'slave_client')
        self._channel_client.start()

    @staticmethod
    def is_context_ready():
        return os.getpid() in _pid_to_context

    @property
    def channel_client(self):
        return self._channel_client

    def register_application(self, app_name, address):
        return self.channel_client.sync_call('report_app_address', json.dumps(
            dict(name=app_name, address=address)
        ))

    def get_bearer_token(self):
        return self.channel_client.sync_call('get_bearer_token', '')

    def kv_store(self):
        """
        Get key-value store of Cupid Service

        :param session: cupid session
        :return: kv-store instance
        """
        from ..io.kvstore import CupidKVStore

        if not hasattr(self, '_cupid_kv_store'):
            self._cupid_kv_store = CupidKVStore()
        return self._cupid_kv_store

    @staticmethod
    def prepare_channel():
        controller = CupidRpcController()
        channel = SandboxRpcChannel()
        stub = process_pb.ProcessService_Stub(channel)
        env = process_pb.EnvEntry()
        resp = stub.Prepare(controller, env, None)
        return resp.entries

    def report_container_status(self, status, message, progress, timeout=-1):
        params = json.dumps(dict(
            status=str(status.value),
            message=json.dumps(dict(
                status=message,
                progress=str(progress.value),
            )),
        ))
        self._channel_client.sync_call("report_container_status", params, timeout=timeout)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from libc.stdint cimport *
from libc.string cimport *
from cpython.bool cimport PyBool_Check
from cpython.datetime cimport import_datetime, PyDateTime_Check
from datetime import datetime

from .. import types, options

cdef int64_t bigint_min = types.bigint._bounds[0]
cdef int64_t bigint_max = types.bigint._bounds[1]
cdef int string_len_max = types.string._max_length
cdef object pd_na_type = types.pd_na_type
cdef bint is_py3 = sys.version_info[0] == 3

cdef:
    int64_t BOOL_TYPE_ID = types.boolean._type_id
    int64_t DATETIME_TYPE_ID = types.datetime._type_id
    int64_t STRING_TYPE_ID = types.string._type_id
    int64_t FLOAT_TYPE_ID = types.float_._type_id
    int64_t DOUBLE_TYPE_ID = types.double._type_id
    int64_t BIGINT_TYPE_ID = types.bigint._type_id
    int64_t BINARY_TYPE_ID = types.binary._type_id
    int64_t TIMESTAMP_TYPE_ID = types.timestamp._type_id
    int64_t JSON_TYPE_ID = types.Json._type_id


import_datetime()


cdef object _validate_bigint(object val, int64_t max_field_size):
    cdef int64_t i_val = val
    if bigint_min <= i_val <= bigint_max:
        return i_val
    raise ValueError('InvalidData: Bigint(%s) out of range' % val)


cdef object _validate_string(object val, int64_t max_field_size):
    cdef:
        size_t s_size
        unicode u_val

    if max_field_size == 0:
        max_field_size = string_len_max

    if isinstance(val, bytes):
        s_size = len(<bytes> val)
        u_val = (<bytes> val).decode('utf-8')
    elif isinstance(val, unicode):
        u_val = <unicode> val
        s_size = 4 * len(u_val)
        if s_size > max_field_size:
            # only encode when strings are long enough
            s_size = len(u_val.encode('utf-8'))
    else:
        raise TypeError("Invalid data type: expect bytes or unicode, got %s" % type(val))

    if s_size <= max_field_size:
        return u_val
    raise ValueError(
        "InvalidData: Length of string(%s) is more than %sM.'" %
        (val, max_field_size / (1024 ** 2))
    )


cdef object _validate_binary(object val, int64_t max_field_size):
    cdef:
        size_t s_size
        bytes bytes_val

    if max_field_size == 0:
        max_field_size = string_len_max

    if isinstance(val, bytes):
        bytes_val = <bytes> val
    elif isinstance(val, unicode):
        bytes_val = (<unicode> val).encode('utf-8')
    else:
        raise TypeError("Invalid data type: expect bytes or unicode, got %s" % type(val))

    s_size = len(bytes_val)
    if s_size <= max_field_size:
        return bytes_val
    raise ValueError(
        "InvalidData: Length of string(%s) is more than %sM.'" %
        (val, max_field_size / (1024 ** 2)))


py_strptime = datetime.strptime

cdef object _validate_datetime(object val, int64_t max_field_size):
    if PyDateTime_Check(val):
        return val
    if isinstance(val, (bytes, unicode)):
        return py_strptime(val, '%Y-%m-%d %H:%M:%S')
    raise TypeError("Invalid data type: expect datetime, got %s" % type(val))


pd_ts = None
pd_ts_strptime = None

cdef object _validate_timestamp(object val, int64_t max_field_size):
    global pd_ts, pd_ts_strptime
    if pd_ts is None:
        try:
            import pandas as pd
            pd_ts = pd.Timestamp
            pd_ts_strptime = pd_ts.strptime
        except ImportError:
            raise ImportError('To use TIMESTAMP in pyodps, you need to install pandas.')

    if isinstance(val, pd_ts):
        return val
    if isinstance(val, (bytes, unicode)):
        return pd_ts_strptime(val, '%Y-%m-%d %H:%M:%S')
    raise TypeError("Invalid data type: expect timestamp, got %s" % type(val))


cdef object _validate_json(object val, int64_t max_field_size):
    if not isinstance(val, (list, dict, unicode, bytes, float, int, long)):
        raise ValueError("Invalid data type: cannot accept %r for json type" % val)

    if is_py3 and type(val) is bytes:
        val = (<bytes> val).decode("utf-8")
    elif not is_py3 and type(val) is unicode:
        val = (<unicode> val).encode("utf-8")
    return val


cdef object _validate_float(object val, int64_t max_field_size):
    cdef float flt_val = val
    return flt_val


cdef object _validate_double(object val, int64_t max_field_size):
    cdef double dbl_val = val
    return dbl_val


cdef object _validate_boolean(object val, int64_t max_field_size):
    if PyBool_Check(val):
        return val
    raise TypeError("Invalid data type: expect bool, got %s" % type(val))


cdef object validate_value(object val, object value_type, int64_t max_field_size):
    if value_type.nullable and (val is None or type(val) is pd_na_type):
        return None

    cdef int64_t type_id = value_type._type_id

    try:
        if type_id == BIGINT_TYPE_ID:
            return _validate_bigint(val, max_field_size)
        elif type_id == FLOAT_TYPE_ID:
            return _validate_float(val, max_field_size)
        elif type_id == DOUBLE_TYPE_ID:
            return _validate_double(val, max_field_size)
        elif type_id == STRING_TYPE_ID:
            if options.tunnel.string_as_binary:
                return _validate_binary(val, max_field_size)
            else:
                return _validate_string(val, max_field_size)
        elif type_id == DATETIME_TYPE_ID:
            return _validate_datetime(val, max_field_size)
        elif type_id == BOOL_TYPE_ID:
            return _validate_boolean(val, max_field_size)
        elif type_id == BINARY_TYPE_ID:
            return _validate_binary(val, max_field_size)
        elif type_id == TIMESTAMP_TYPE_ID:
            return _validate_timestamp(val, max_field_size)
        elif type_id == JSON_TYPE_ID:
            return _validate_json(val, max_field_size)
    except TypeError:
        pass

    return types.validate_value(val, value_type, max_field_size=max_field_size)


cdef class SchemaSnapshot:
    def __cinit__(self, schema):
        cdef int type_id, i
        self._columns = schema.columns
        self._col_count = len(schema)
        self._col_types = [c.type for c in schema.columns]
        self._col_type_ids = [t._type_id if t._type_id is not None else -1 for t in self._col_types]
        self._col_nullable = [t.nullable for t in self._col_types]
        if isinstance(schema, types.OdpsSchema):
            self._col_is_partition = [
                1 if schema.is_partition(c) else 0
                for c in schema.columns
            ]
            self._partition_col_count = sum(self._col_is_partition)
        else:
            self._col_is_partition = [0] * self._col_count
            self._partition_col_count = 0

        self._col_validators.resize(self._col_count)
        for i in range(self._col_count):
            type_id = self._col_type_ids[i]
            if type_id == BIGINT_TYPE_ID:
                self._col_validators[i] = _validate_bigint
            elif type_id == FLOAT_TYPE_ID:
                self._col_validators[i] = _validate_float
            elif type_id == DOUBLE_TYPE_ID:
                self._col_validators[i] = _validate_double
            elif type_id == STRING_TYPE_ID:
                if options.tunnel.string_as_binary:
                    self._col_validators[i] = _validate_binary
                else:
                    self._col_validators[i] = _validate_string
            elif type_id == DATETIME_TYPE_ID:
                self._col_validators[i] = _validate_datetime
            elif type_id == BOOL_TYPE_ID:
                self._col_validators[i] = _validate_boolean
            elif type_id == BINARY_TYPE_ID:
                self._col_validators[i] = _validate_binary
            elif type_id == TIMESTAMP_TYPE_ID:
                self._col_validators[i] = _validate_timestamp
            elif type_id == JSON_TYPE_ID:
                self._col_validators[i] = _validate_json
            else:
                self._col_validators[i] = NULL

    cdef object validate_value(self, int i, object val, int64_t max_field_size):
        cdef _VALIDATE_FUNC vfun = self._col_validators[i]
        if val is None and self._col_nullable[i]:
            return val
        if vfun != NULL:
            try:
                return vfun(val, max_field_size)
            except TypeError:
                pass
        return types.validate_value(val, self._col_types[i], max_field_size=max_field_size)


cdef class BaseRecord:
    def __cinit__(self, columns=None, schema=None, values=None, max_field_size=None):
        self._c_schema_snapshot = getattr(schema, '_snapshot', None)
        if columns is not None:
            self._c_columns = columns
            self._c_name_indexes = {col.name: i for i, col in enumerate(self._c_columns)}
        else:
            self._c_columns = schema.columns if self._c_schema_snapshot is None else self._c_schema_snapshot._columns
            self._c_name_indexes = schema._name_indexes

        self._max_field_size = max_field_size or 0

        if self._c_columns is None:
            raise ValueError('Either columns or schema should not be provided')

        self._c_values = [None] * len(self._c_columns)
        if values is not None:
            self._sets(values)

    def __reduce__(self):
        return type(self), (self._c_columns, None, self._c_values)

    @property
    def _columns(self):
        return self._c_columns

    @_columns.setter
    def _columns(self, value):
        self._c_columns = value

    @property
    def _values(self):
        return self._c_values

    @_values.setter
    def _values(self, value):
        self._c_values = value

    @property
    def _name_indexes(self):
        return self._c_name_indexes

    @_name_indexes.setter
    def _name_indexes(self, value):
        self._c_name_indexes = value

    def _mode(self):
        return 'c'

    cdef size_t _get_non_partition_col_count(self):
        if self._c_schema_snapshot is not None:
            return self._c_schema_snapshot._col_count - self._c_schema_snapshot._partition_col_count
        else:
            return len([col for col in self._c_columns if not isinstance(col, types.Partition)])

    cpdef object get_by_name(self, object name):
        cdef int idx = self._c_name_indexes[name]
        return self._c_values[idx]

    cpdef set_by_name(self, object name, object value):
        cdef int idx = self._c_name_indexes[name]
        self._set(idx, value)

    cpdef _set(self, int i, object value):
        cdef object t
        if self._c_schema_snapshot is not None:
            value = self._c_schema_snapshot.validate_value(i, value, self._max_field_size)
        else:
            t = self._c_columns[i].type
            try:
                value = validate_value(value, t, self._max_field_size)
            except TypeError:
                value = types.validate_value(
                    value, t, max_field_size=self._max_field_size
                )

        self._c_values[i] = value

    cpdef object _get(self, int i):
        return self._c_values[i]

    get = _get  # to keep compatible
    set = _set  # to keep compatible

    cpdef _sets(self, object values):
        cdef int i
        cdef object value
        cdef size_t n_values

        if type(values) is list:
            n_values = len(<list>values)
        else:
            n_values = len(values)

        if (
            n_values != len(self._c_columns) and
            n_values != self._get_non_partition_col_count()
        ):
            raise ValueError(
                'The values set to records are against the schema, '
                'expect len %s, got len %s' % (len(self._c_columns), n_values)
            )

        if type(values) is list:
            for i, value in enumerate(<list>values):
                self._set(i, value)
        else:
            for i, value in enumerate(values):
                self._set(i, value)

    def __getitem__(self, item):
        if isinstance(item, (bytes, unicode)):
            return self.get_by_name(item)
        elif isinstance(item, (list, tuple)):
            return [self[it] for it in item]
        return self._values[item]

    def __setitem__(self, key, value):
        if isinstance(key, (bytes, unicode)):
            self.set_by_name(key, value)
        else:
            self._set(key, value)

    def __getattr__(self, item):
        if item == '_name_indexes':
            return self._c_name_indexes
        if item in self._c_name_indexes:
            i = self._c_name_indexes[item]
            return self._c_values[i]
        return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        cdef int i
        if key in self._c_name_indexes:
            i = self._c_name_indexes[key]
            self._set(i, value)
        else:
            object.__setattr__(self, key, value)

    def __len__(self):
        return len(self._c_columns)

    def __contains__(self, item):
        return item in self._c_name_indexes

    def __iter__(self):
        cdef int i
        cdef object col
        for i, col in enumerate(self._c_columns):
            yield (col.name, self._get(i))

    @property
    def values(self):
        return self._c_values

    @property
    def n_columns(self):
        if self._c_schema_snapshot is not None:
            return self._c_schema_snapshot._col_count
        return len(self._c_columns)

    def get_columns_count(self):  # compatible
        return self.n_columns



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import logging
import time
import requests
import warnings

from mars import new_session

from ....models import Instance
from ....config import options
from .... import errors
from ...utils import get_default_resource_files, build_mars_image_name

NOTEBOOK_NAME = "MarsNotebook"
CUPID_APP_NAME = "MarsWeb"
GS_COORDINATOR_NAME = "GSCoordinator"
DEFAULT_RESOURCES = ["pymars-0.9.0", "pyodps-0.11.5", "pyarrow-4.0.0"]

logger = logging.getLogger(__name__)


class MarsCupidClient:
    def __init__(self, odps, inst=None, project=None):
        from cupid import CupidSession

        self._odps = odps
        self._cupid_session = CupidSession(odps, project=project)
        self._kube_instance = inst
        self._kube_url = None
        self._kube_client = None
        self._kube_namespace = None

        self._supervisor_key = None
        self._supervisor_config = None
        self._worker_config = None
        self._endpoint = None
        self._with_notebook = False
        self._notebook_endpoint = None
        self._with_graphscope = False
        self._graphscope_endpoint = None

        self._mars_session = None
        self._req_session = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def notebook_endpoint(self):
        return self._notebook_endpoint

    @property
    def session(self):
        return self._mars_session

    @property
    def instance_id(self):
        return self._kube_instance.id

    def rewrite_mars_request(self, request):
        from ....rest import default_user_agent

        request.headers["User-Agent"] = request.user_agent = default_user_agent()
        if "Content-Type" not in request.headers:
            request.headers["Content-Type"] = "application/octet-stream"

        if options.mars.use_common_proxy:
            if self._odps.project is not None:
                project_query = "curr_project=" + self._odps.project
                if "?" in request.url:
                    request.url += "&" + project_query
                else:
                    request.url += "?" + project_query
            account = self._req_session.account
            account.sign_request(request, self._endpoint)
        return request

    def submit(
        self,
        image=None,
        supervisor_num=1,
        supervisor_cpu=4,
        supervisor_mem=16 * 1024**3,
        worker_num=1,
        worker_cpu=8,
        worker_mem=32 * 1024**3,
        worker_cache_mem=None,
        min_worker_num=None,
        worker_disk_num=1,
        worker_disk_size=100 * 1024**3,
        with_notebook=False,
        notebook_cpu=1,
        notebook_mem=2 * 1024**3,
        with_graphscope=False,
        coordinator_cpu=1,
        coordinator_mem=2 * 1024**3,
        timeout=None,
        extra_env=None,
        extra_modules=None,
        resources=None,
        create_session=True,
        priority=None,
        running_cluster=None,
        task_name=None,
        **kw
    ):
        try:
            async_ = kw.pop("async_", None)

            # compatible with early version
            mars_image = kw.pop("mars_image", None)
            default_resources = kw.pop(
                "default_resources", None
            ) or get_default_resource_files(DEFAULT_RESOURCES)
            instance_idle_timeout = kw.pop("instance_idle_timeout", None)
            node_blacklist = kw.pop("node_blacklist", None)
            if with_notebook is not None:
                self._with_notebook = bool(with_notebook)
            else:
                self._with_notebook = options.mars.launch_notebook
            if with_graphscope is not None:
                self._with_graphscope = bool(with_graphscope)
            if self._kube_instance is None:
                image = image or build_mars_image_name(mars_image)

                extra_modules = extra_modules or []
                if isinstance(extra_modules, (tuple, list)):
                    extra_modules = list(extra_modules) + ["odps.mars_extension"]
                else:
                    extra_modules = [extra_modules, "odps.mars_extension"]

                if resources is not None:
                    if isinstance(resources, (tuple, list)):
                        resources = list(resources)
                        resources.extend(default_resources)
                    else:
                        resources = [resources] + default_resources
                else:
                    resources = default_resources

                if worker_cache_mem is None:
                    worker_cache_mem = int(worker_mem * 0.4)
                else:
                    worker_cache_mem = worker_cache_mem

                cluster_args = dict(
                    image=image,
                    supervisor_num=supervisor_num,
                    supervisor_cpu=supervisor_cpu,
                    supervisor_mem=supervisor_mem,
                    worker_num=worker_num,
                    worker_cpu=worker_cpu,
                    worker_mem=worker_mem,
                    worker_cache_mem=worker_cache_mem,
                    min_worker_num=min_worker_num,
                    worker_disk_num=worker_disk_num,
                    worker_disk_size=worker_disk_size,
                    with_notebook=with_notebook,
                    notebook_cpu=notebook_cpu,
                    notebook_mem=notebook_mem,
                    with_graphscope=with_graphscope,
                    coordinator_cpu=coordinator_cpu,
                    coordinator_mem=coordinator_mem,
                    extra_env=extra_env,
                    extra_modules=extra_modules,
                    node_blacklist=node_blacklist,
                    instance_idle_timeout=instance_idle_timeout,
                    timeout=timeout,
                )

                command = "/srv/entrypoint.sh %s %s" % (
                    __name__.rsplit(".", 1)[0] + ".app",
                    base64.b64encode(json.dumps(cluster_args).encode()).decode(),
                )

                self._kube_instance = self._cupid_session.start_kubernetes(
                    async_=True,
                    running_cluster=running_cluster,
                    priority=priority,
                    app_image=build_mars_image_name(),
                    app_command=command,
                    resources=resources,
                    task_name=task_name,
                    **kw
                )
                logger.info(self._kube_instance.get_logview_address())
            if async_:
                return self
            else:
                self.wait_for_success(
                    create_session=create_session,
                    min_worker_num=min_worker_num or worker_num,
                )
                return self

        except KeyboardInterrupt:
            self.stop_server()
            return self

    def check_service_ready(self, timeout=1):
        try:
            resp = self._req_session.get(self._endpoint + "/api", timeout=timeout)
        except (requests.ConnectionError, requests.Timeout, errors.ODPSError):
            return False
        if resp.status_code >= 400:
            return False
        elif b"not found" in resp.content:
            self._endpoint = None
            return False
        return True

    def count_workers(self):
        resp = self._req_session.get(
            self._endpoint + "/api/cluster/nodes?role=1", timeout=1
        )
        json_result = json.loads(resp.text)
        return len(json_result["nodes"])

    def rescale_workers(self, new_scale, min_workers=None, wait=True, timeout=None):
        self._mars_session._sess.rescale_workers(
            new_scale, min_workers=min_workers, wait=wait, timeout=timeout
        )

    def get_logview_address(self):
        return self._kube_instance.get_logview_address()

    def get_mars_endpoint(self):
        return self._cupid_session.get_proxied_url(
            self._kube_instance.id, CUPID_APP_NAME
        )

    def get_notebook_endpoint(self):
        return self._cupid_session.get_proxied_url(
            self._kube_instance.id, NOTEBOOK_NAME
        )

    def get_graphscope_endpoint(self):
        return self._cupid_session.get_proxied_url(
            self._kube_instance.id, GS_COORDINATOR_NAME
        )

    def get_req_session(self):
        from ....rest import RestClient

        if options.mars.use_common_proxy:
            return RestClient(
                self._odps.account, self._endpoint, self._odps.project, tag="MARS"
            )
        else:
            return requests.Session()

    def check_instance_status(self):
        if self._kube_instance.is_terminated():
            for task_name, task in (self._kube_instance.get_task_statuses()).items():
                exc = None
                if task.status == Instance.Task.TaskStatus.FAILED:
                    exc = errors.parse_instance_error(
                        self._kube_instance.get_task_result(task_name)
                    )
                elif task.status != Instance.Task.TaskStatus.SUCCESS:
                    exc = errors.ODPSError(
                        "%s, status=%s" % (task_name, task.status.value)
                    )
                if exc:
                    exc.instance_id = self._kube_instance.id
                    raise exc

    def _post_pyodps_api(self, **data):
        r = self._req_session.post(
            self._endpoint.rstrip("/") + "/api/pyodps", data=data
        )
        try:
            r.raise_for_status()
        except errors.InvalidStateSetting:
            if not self._kube_instance.is_successful():
                raise

    def wait_for_success(self, min_worker_num=0, create_session=True):
        while True:
            self.check_instance_status()
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")
                try:
                    if self._endpoint is None:
                        self._endpoint = self.get_mars_endpoint()
                        logger.info("Mars UI: " + self._endpoint)
                        self._req_session = self.get_req_session()

                        self._post_pyodps_api(
                            action="write_log",
                            content="Mars UI from client: " + self._endpoint,
                        )
                    if self._with_notebook and self._notebook_endpoint is None:
                        self._notebook_endpoint = self.get_notebook_endpoint()
                        logger.info("Notebook UI: " + self._notebook_endpoint)

                        self._post_pyodps_api(
                            action="write_log",
                            content="Notebook UI from client: "
                            + self._notebook_endpoint,
                        )
                    if self._with_graphscope and self._graphscope_endpoint is None:
                        self._graphscope_endpoint = self.get_graphscope_endpoint()
                        logger.info("Graphscope endpoint: " + self._graphscope_endpoint)

                        self._post_pyodps_api(
                            action="write_log",
                            content="Graphscope endpoint from client: "
                            + self._graphscope_endpoint,
                        )
                except KeyboardInterrupt:
                    raise
                except:
                    time.sleep(1)
                    continue

                try:
                    if not self.check_service_ready():
                        continue
                    if self.count_workers() >= min_worker_num:
                        break
                except:
                    continue
                finally:
                    time.sleep(1)

        if create_session:
            try:
                self._mars_session = new_session(
                    self._endpoint, request_rewriter=self.rewrite_mars_request
                ).as_default()
            except KeyboardInterrupt:
                raise
            except:
                if (
                    self._kube_instance
                    and self._kube_instance.status == self._kube_instance.Status.RUNNING
                ):
                    self._kube_instance.stop()
                raise

    def restart_session(self):
        self._mars_session.close()
        self._mars_session = new_session(
            self._endpoint, request_rewriter=self.rewrite_mars_request
        ).as_default()

    def stop_server(self):
        if not self._kube_instance:
            return

        try:
            self._post_pyodps_api(action="terminate", message="Stopped at client side")
            self._kube_instance.wait_for_completion(
                timeout=options.mars.container_status_timeout
            )
        except BaseException:
            if not self._kube_instance.is_terminated():
                try:
                    self._kube_instance.stop()
                except errors.InvalidStateSetting:
                    pass
        finally:
            self._kube_instance = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict, OrderedDict

from .. import serializers, errors, types
from ..compat import six
from ..utils import with_wait_argument
from .core import Iterable
from .partition import Partition


class PartitionSpecCondition(object):
    _predicates = OrderedDict([
        ("==", lambda a, b: a == b),
        (">=", lambda a, b: a >= b),
        ("<=", lambda a, b: a <= b),
        ("<>", lambda a, b: a != b),
        ("!=", lambda a, b: a != b),
        (">", lambda a, b: a > b),
        ("<", lambda a, b: a < b),
        ("=", lambda a, b: a == b),
    ])

    def __init__(self, part_fields, condition=None):
        self._part_to_conditions = defaultdict(list)
        field_set = set(part_fields)
        condition = str(condition) if condition else None
        condition_splits = condition.split(",") if condition else []
        for split in condition_splits:
            for pred in self._predicates:
                if pred not in split:
                    continue

                parts = split.split(pred, 1)
                if len(parts) != 2:
                    raise ValueError("Invalid partition condition %r" % split)
                part = parts[0].strip()
                val = parts[1].strip().replace('"', '').replace("'", '')

                if part not in field_set:
                    raise ValueError("Invalid partition field %r" % part)

                self._part_to_conditions[part].append((pred, val))
                break
            else:
                raise ValueError("Invalid partition condition %r" % split)
        specs = []
        for field in part_fields:
            if (
                field not in self._part_to_conditions
                or len(self._part_to_conditions[field]) > 1
                or self._part_to_conditions[field][0][0] not in ("=", "==")
            ):
                break
            specs.append("%s=%s" % (field, self._part_to_conditions.pop(field)[0][1]))
        self.partition_spec = types.PartitionSpec(",".join(specs)) if specs else None

    def match(self, spec):
        for field, conditions in self._part_to_conditions.items():
            real_val = spec[field]
            for pred, val in conditions:
                if not self._predicates[pred](real_val, val):
                    return False
        return True


class Partitions(Iterable):
    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    partitions = serializers.XMLNodesReferencesField(Partition, 'Partition')

    def _name(self):
        return

    def _get(self, item):
        return Partition(client=self._client, parent=self, spec=item)

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            item = types.PartitionSpec(item)
            return self._get(item)
        elif isinstance(item, types.PartitionSpec):
            return self._get(item)
        return super(Partitions, self).__getitem__(item)

    def __contains__(self, item):
        if isinstance(item, (six.string_types, types.PartitionSpec)):
            if isinstance(item, six.string_types):
                item = types.PartitionSpec(item)
            partition = self._get(item)
        elif isinstance(item, Partition):
            partition = item
        else:
            return False

        try:
            partition.reload()
            return True
        except errors.NoSuchObject:
            return False

    @classmethod
    def _get_partition_spec(cls, partition_spec):
        if isinstance(partition_spec, types.PartitionSpec):
            return partition_spec
        return types.PartitionSpec(partition_spec)

    def __iter__(self):
        return self.iterate_partitions()

    @property
    def project(self):
        return self.parent.project

    def iterate_partitions(self, spec=None, reverse=False):
        condition = PartitionSpecCondition(
            [pt.name for pt in self.parent.table_schema.partitions], spec
        )
        spec = condition.partition_spec

        actions = ['partitions']
        params = {'expectmarker': 'true'}
        if reverse:
            actions.append('reverse')
        if spec is not None and not spec.is_empty:
            params['partition'] = str(spec)
        schema_name = self._get_schema_name()
        if schema_name:
            params['curr_schema'] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, actions=actions, params=params)

            t = self.parse(self._client, resp, obj=self)
            params['marker'] = t.marker

            return t.partitions

        while True:
            partitions = _it()
            if partitions is None:
                break
            for partition in partitions:
                if condition.match(partition.partition_spec):
                    yield partition

    def get_max_partition(self, spec=None, skip_empty=True, reverse=False):
        table_parts = self.parent.table_schema.partitions

        if spec is not None:
            spec = self._get_partition_spec(spec)
            if len(spec) >= len(table_parts):
                raise ValueError(
                    "Size of prefix should not exceed number of partitions of the table"
                )

            for exist_pt, user_pt_name in zip(table_parts, spec.kv):
                if exist_pt.name != user_pt_name:
                    table_pt_str = ",".join(pt.name for pt in table_parts[:len(spec)])
                    prefix_pt_str = ",".join(spec.kv.keys())
                    raise ValueError(
                        "Partition prefix %s not agree with table partitions %s",
                        prefix_pt_str,
                        table_pt_str,
                    )

        part_values = [
            (part, tuple(part.partition_spec.values()))
            for part in self.iterate_partitions(spec)
        ]
        if not part_values:
            return None
        elif not skip_empty:
            return max(part_values, key=lambda tp: tp[1])[0]
        else:
            reversed_table_parts = sorted(part_values, key=lambda tp: tp[1], reverse=not reverse)
            return next(
                (
                    part
                    for part, _ in reversed_table_parts
                    if not skip_empty or part.physical_size > 0
                ),
                None,
            )

    @with_wait_argument
    def create(self, partition_spec, if_not_exists=False, async_=False, hints=None):
        if isinstance(partition_spec, Partition):
            partition_spec = partition_spec.partition_spec
        else:
            partition_spec = self._get_partition_spec(partition_spec)

        buf = six.StringIO()
        buf.write('ALTER TABLE %s ADD ' % self.parent.full_table_name)

        if if_not_exists:
            buf.write('IF NOT EXISTS ')

        buf.write('PARTITION (%s);' % partition_spec)

        from .tasks import SQLTask
        task = SQLTask(name='SQLAddPartitionTask', query=buf.getvalue())
        hints = hints or {}
        schema_name = self._get_schema_name()
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        task.update_sql_settings(hints)
        instance = self.project.parent[self._client.project].instances.create(task=task)

        if not async_:
            instance.wait_for_success()
            return self[partition_spec]
        else:
            return instance

    @with_wait_argument
    def delete(self, partition_spec, if_exists=False, async_=False, hints=None):
        if isinstance(partition_spec, Partition):
            partition_spec = partition_spec.partition_spec
        else:
            partition_spec = self._get_partition_spec(partition_spec)

        buf = six.StringIO()
        buf.write('ALTER TABLE %s DROP ' % self.parent.full_table_name)

        if if_exists:
            buf.write('IF EXISTS ')

        buf.write('PARTITION (%s);' % partition_spec)

        from .tasks import SQLTask
        task = SQLTask(name='SQLDropPartitionTask', query=buf.getvalue())

        hints = hints or {}
        hints['odps.sql.submit.mode'] = ''
        schema_name = self._get_schema_name()
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        task.update_sql_settings(hints)
        instance = self.project.parent[self._client.project].instances.create(task=task)

        if not async_:
            instance.wait_for_success()
        else:
            return instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified from part of python-hashes by sangelone.


import math
import hashlib
import sys
try:
    import cPickle as pickle
except ImportError:
    import pickle

if sys.version < '3':
    PY3 = False
else:
    PY3 = True


default_hashbits = 96


class HashType(object):
    def __init__(self, value='', hashbits=default_hashbits, hash_=None):
        "Relies on create_hash() provided by subclass"
        self.hashbits = hashbits
        if hash_:
            self.hash = hash_
        else:
            self.create_hash(value)

    def __trunc__(self):
        return self.hash

    def __str__(self):
        return str(self.hash)

    def __int__(self):
        return int(self.hash)

    def __float__(self):
        return float(self.hash)

    def __cmp__(self, other):
        if self.hash < int(other): return -1
        if self.hash > int(other): return 1
        return 0

    def hex(self):
        return hex(self.hash)

    def hamming_distance(self, other_hash):
        x = (self.hash ^ other_hash.hash) & ((1 << self.hashbits) - 1)
        tot = 0
        while x:
            tot += 1
            x &= x-1
        return tot


class BloomFilter(HashType):
    def __init__(self, capacity=3000, false_positive_rate=0.01, value=''):
        """
        'value' is the initial string or list of strings to hash,
        'capacity' is the expected upper limit on items inserted, and
        'false_positive_rate' is self-explanatory but the smaller it is, the larger your hashes!
        """
        self.capacity = capacity
        self.error_rate = false_positive_rate
        self.create_hash(value, capacity, false_positive_rate)

    def create_hash(self, initial, capacity, error):
        """
        Calculates a Bloom filter with the specified parameters.
        Initalizes with a string or list/set/tuple of strings. No output.

        Reference material: http://bitworking.org/news/380/bloom-filter-resources
        """
        self.hash = 0
        self.hashbits, self.num_hashes = self._optimal_size(capacity, error)

        if len(initial):
            if type(initial) == str:
                self.add(initial)
            else:
                for t in initial:
                    self.add(t)

    def _binary(self, item):
        if PY3 and isinstance(item, str):
            item = item.encode('utf-8')

        return item

    def _hashes(self, item):
        """
        To create the hash functions we use the SHA-1 hash of the
        string and chop that up into 20 bit values and then
        mod down to the length of the Bloom filter.
        """
        item = self._binary(item)

        m = hashlib.sha1()
        m.update(item)
        digits = m.hexdigest()

        # Add another 160 bits for every 8 (20-bit long) hashes we need
        for i in range(int(self.num_hashes // 8)):
            m.update(self._binary(str(i)))
            digits += m.hexdigest()

        hashes = [int(digits[i*5:i*5+5], 16) % self.hashbits for i in range(self.num_hashes)]
        return hashes

    def _optimal_size(self, capacity, error):
        """Calculates minimum number of bits in filter array and
        number of hash functions given a number of enteries (maximum)
        and the desired error rate (falese positives).

        Example:
            m, k = self._optimal_size(3000, 0.01)   # m=28756, k=7
        """
        m = math.ceil((capacity * math.log(error)) / math.log(1.0 / (math.pow(2.0, math.log(2.0)))))
        k = math.ceil(math.log(2.0) * m / capacity)
        return int(m), int(k)

    def add(self, item):
        "Add an item (string) to the filter. Cannot be removed later!"
        for pos in self._hashes(item):
            self.hash |= (2 ** pos)

    def union(self, other):
        if self.capacity != other.capacity or self.error_rate != other.error_rate:
            raise ValueError('Unioning filters requires both filters to have both '
                             'the same capacity and error rate')

        new_bloom = BloomFilter(capacity=self.capacity,
                                false_positive_rate=self.error_rate)
        new_bloom.hash = self.hash
        new_bloom.hash |= other.hash
        return new_bloom

    def intersection(self, other):
        if self.capacity != other.capacity or self.error_rate != other.error_rate:
            raise ValueError('Intersecting filters requires both filters to have both '
                             'the same capacity and error rate')

        new_bloom = BloomFilter(capacity=self.capacity,
                                false_positive_rate=self.error_rate)
        new_bloom.hash = self.hash
        new_bloom.hash &= other.hash
        return new_bloom

    def __contains__(self, name):
        "This function is used by the 'in' keyword"
        retval = True
        for pos in self._hashes(name):
            retval = retval and bool(self.hash & (2 ** pos))
        return retval



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""

import sys
import array
import struct

from . import errors
from . import wire_format


class OutputStream(object):
    """Contains all logic for writing bits, and ToString() to get the result."""

    def __init__(self):
        self._buffer = array.array('B')

    if sys.version_info < (3, 3):
        def append_raw_bytes(self, raw_bytes):
            """Appends raw_bytes to our internal buffer."""
            self._buffer.fromstring(raw_bytes)
    else:
        def append_raw_bytes(self, raw_bytes):
            """Appends raw_bytes to our internal buffer."""
            self._buffer.frombytes(raw_bytes)

    def append_little_endian32(self, unsigned_value):
        """Appends an unsigned 32-bit integer to the internal buffer,
        in little-endian byte order.
        """
        if not 0 <= unsigned_value <= wire_format.UINT32_MAX:
            raise errors.EncodeError(
                'Unsigned 32-bit out of range: %d' % unsigned_value)
        self.append_raw_bytes(struct.pack(
            wire_format.FORMAT_UINT32_LITTLE_ENDIAN, unsigned_value))

    def append_little_endian64(self, unsigned_value):
        """Appends an unsigned 64-bit integer to the internal buffer,
        in little-endian byte order.
        """
        if not 0 <= unsigned_value <= wire_format.UINT64_MAX:
            raise errors.EncodeError(
                'Unsigned 64-bit out of range: %d' % unsigned_value)
        self.append_raw_bytes(struct.pack(
            wire_format.FORMAT_UINT64_LITTLE_ENDIAN, unsigned_value))

    def append_varint32(self, value):
        """Appends a signed 32-bit integer to the internal buffer,
        encoded as a varint.  (Note that a negative varint32 will
        always require 10 bytes of space.)
        """
        if not wire_format.INT32_MIN <= value <= wire_format.INT32_MAX:
            raise errors.EncodeError('Value out of range: %d' % value)
        self.append_varint64(value)

    def append_var_uint32(self, value):
        """Appends an unsigned 32-bit integer to the internal buffer,
        encoded as a varint.
        """
        if not 0 <= value <= wire_format.UINT32_MAX:
            raise errors.EncodeError('Value out of range: %d' % value)
        self.append_var_uint64(value)

    def append_varint64(self, value):
        """Appends a signed 64-bit integer to the internal buffer,
        encoded as a varint.
        """
        if not wire_format.INT64_MIN <= value <= wire_format.INT64_MAX:
            raise errors.EncodeError('Value out of range: %d' % value)
        if value < 0:
            value += (1 << 64)
        self.append_var_uint64(value)

    def append_var_uint64(self, unsigned_value):
        """Appends an unsigned 64-bit integer to the internal buffer,
        encoded as a varint.
        """
        if not 0 <= unsigned_value <= wire_format.UINT64_MAX:
            raise errors.EncodeError('Value out of range: %d' % unsigned_value)
        while True:
            bits = unsigned_value & 0x7f
            unsigned_value >>= 7
            if unsigned_value:
                bits |= 0x80
            self._buffer.append(bits)
            if not unsigned_value:
                break

    if sys.version_info < (3, 3):
        def tostring(self):
            """Returns a string containing the bytes in our internal buffer."""
            return self._buffer.tostring()
    else:
        def tostring(self):
            """Returns a string containing the bytes in our internal buffer."""
            return self._buffer.tobytes()

    def __len__(self):
        return len(self._buffer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import inspect
import os
import platform
import sys
import uuid
from collections import OrderedDict

from .types import df_type_to_odps_type
from ...expr.collections import RowAppliedCollectionExpr
from ...expr.element import MappedExpr
from ...expr.reduction import Aggregation, GroupedAggregation
from ...expr.utils import get_executed_collection_project_table_name
from ...utils import make_copy
from ....config import options
from ....lib import cloudpickle
from ....compat import six
from ....models import FileResource, TableResource, ArchiveResource
from ....utils import to_str

dirname = os.path.dirname(os.path.abspath(cloudpickle.__file__))
CLOUD_PICKLE_FILE = os.path.join(dirname, 'cloudpickle.py')
with open(CLOUD_PICKLE_FILE) as f:
    CLOUD_PICKLE = f.read()

IMPORT_FILE = os.path.join(dirname, 'importer.py')
with open(IMPORT_FILE) as f:
    MEM_IMPORT = f.read()

CLIENT_IMPL = '(%d, %d, "%s")' % (sys.version_info[0],
                                  sys.version_info[1],
                                  platform.python_implementation().lower())

X_NAMED_TUPLE_FILE = os.path.join(dirname, 'xnamedtuple.py')
with open(X_NAMED_TUPLE_FILE) as f:
    X_NAMED_TUPLE = f.read()


UDF_TMPL_HEADER = '''\
%(cloudpickle)s
%(memimport)s

%(xnamedtuple)s

try:
    # workaround for character not in range error
    import _strptime
except:
    pass

import sys
# avoid conflict between protobuf binaries
sys.setdlopenflags(10)

import base64
import inspect
import time
import os
try:
    import faulthandler
    faulthandler.enable(all_threads=True)
except ImportError:
    pass

from odps.udf import annotate
from odps.distcache import get_cache_file, get_cache_table, get_cache_archive

PY2 = sys.version_info[0] == 2

try:
    from odps.distcache import get_cache_archive_filenames
except ImportError:
    def get_cache_archive_filenames(name, relative_path='.'):
        from odps.distcache import WORK_DIR, DistributedCacheError

        def _is_parent(parent, child):
            return parent == child or child.startswith(parent + '/')

        if os.path.split(name)[0] != '':
            raise DistributedCacheError("Invalid resource name: " + name)
        ret_files = []

        # find the real resource path to avoid the symbol link in inner system
        resourcepath = os.path.realpath(os.path.join(WORK_DIR, name))

        # use realpath == abspath to check the symbol link
        dirpath = os.path.join(resourcepath, relative_path)
        if not os.path.exists(dirpath):
            raise DistributedCacheError("Invalid relative path, file not exists: " + relative_path)

        if os.path.realpath(dirpath) != os.path.abspath(dirpath):
            raise DistributedCacheError("Invalid relative path, relative path contains symlink: " + relative_path)

        if not _is_parent(resourcepath, dirpath):
            raise DistributedCacheError("Invalid relative path, path not correct in archive: " + relative_path)

        if not os.path.isdir(dirpath):
            return [dirpath]

        for root, dirs, files in os.walk(dirpath):
            for f in dirs:
                filename = os.path.join(root, f)
                if os.path.islink(filename):
                    relativename = os.path.relpath(filename, resourcepath)
                    raise DistributedCacheError("Invalid relative path, relative path contains symlink: " + relativename)

            for f in files:
                filename = os.path.join(root, f)
                if os.path.islink(filename):
                    relativename = os.path.relpath(filename, resourcepath)
                    raise DistributedCacheError("Invalid relative path, relative path contains symlink: " + relativename)
                ret_files.append(filename)
        return ret_files


def get_cache_archive_data(name, relative_path='.'):
    try:
        return [os.path.normpath(f) for f in get_cache_archive_filenames(name, relative_path)]
    except RuntimeError:
        return {os.path.normpath(fo.name): fo for fo in get_cache_archive(name, relative_path)}


class UnbufferedStream(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        if attr != 'stream':
            return getattr(self.stream, attr)
        else:
            return object.__getattribute__(self, 'stream')

    def __setattr__(self, attr, value):
        if attr != 'stream':
            return setattr(self.stream, attr, value)
        else:
            return object.__setattr__(self, 'stream', value)

sys.stdout = UnbufferedStream(sys.stdout)


try:
    import socket
except ImportError:
    class MockSocketModule(object):
        _GLOBAL_DEFAULT_TIMEOUT = object()
        def __getattr__(self, item):
            raise AttributeError('Accessing attribute `{0}` of module `socket` is prohibited by sandbox.'.format(item))
    sys.modules['socket'] = MockSocketModule()


def gen_resource_data(fields, tb):
    named_args = xnamedtuple('NamedArgs', fields)
    for args in tb:
        yield named_args(*args)


def read_lib(lib, f):
    if isinstance(f, (list, dict)):
        return f
    if lib.endswith('.zip') or lib.endswith('.egg') or lib.endswith('.whl'):
        if PY2 and hasattr(f, "read"):
            return zipfile.ZipFile(f)
        else:
            return zipfile.ZipFile(f.name)
    if lib.endswith('.tar') or lib.endswith('.tar.gz') or lib.endswith('.tar.bz2'):
        from io import BytesIO
        if lib.endswith('.tar'):
            mode = 'r'
        else:
            mode = 'r:gz' if lib.endswith('.tar.gz') else 'r:bz2'
        if PY2 and hasattr(f, "read"):
            return tarfile.open(name='', fileobj=f, mode=mode)
        else:
            return tarfile.open(name=f.name, mode=mode)
    if lib.endswith('.py'):
        if PY2 and hasattr(f, "read"):
            return {f.name: f}
        else:
            return {f.name: open(f.name, 'rb')}

    raise ValueError(
        'Unknown library type which should be one of zip(egg, wheel), tar, or tar.gz')


# Use this method to make testThirdPartyLibraries happy
np_generic = None
def load_np_generic():
    global np_generic

    try:
        from numpy import generic
        np_generic = generic
    except ImportError:
        class PseudoNpGeneric(object):
            pass
        np_generic = PseudoNpGeneric
''' % {
    'cloudpickle': CLOUD_PICKLE,
    'memimport': MEM_IMPORT,
    'xnamedtuple': X_NAMED_TUPLE,
}


UDF_TMPL = '''

@annotate('%(from_type)s->%(to_type)s')
class %(func_cls_name)s(object):
    def __init__(self):
        unpickler_kw = dict(impl=%(implementation)s, dump_code=%(dump_code)s)
        rs = loads(base64.b64decode('%(resources)s'), **unpickler_kw)
        resources = []
        for t, n, fields in rs:
            if t == 'file':
                resources.append(get_cache_file(str(n)))
            elif t == 'archive':
                resources.append(get_cache_archive(str(n)))
            else:
                tb = get_cache_table(str(n))
                if fields:
                    tb = gen_resource_data(fields, tb)
                resources.append(tb)

        libraries = (l for l in '%(libraries)s'.split(',') if len(l) > 0)
        files = []
        for lib in libraries:
            if lib.startswith('a:'):
                lib = lib[2:]
                f = get_cache_archive_data(lib)
            else:
                f = get_cache_file(lib)
            files.append(read_lib(lib, f))
        sys.meta_path.append(CompressImporter(*files, supersede=%(supersede_libraries)r))

        load_np_generic()

        encoded = '%(func_str)s'
        f_str = base64.b64decode(encoded)
        self.f = loads(f_str, **unpickler_kw)

        if inspect.isclass(self.f):
            if resources:
                self.f = self.f(resources)
            else:
                self.f = self.f()
        else:
            if resources:
                self.f = self.f(resources)

        self.names = tuple(it for it in '%(names_str)s'.split(',') if it)
        if self.names:
            self.named_args = xnamedtuple('NamedArgs', self.names)

        encoded_func_args = '%(func_args_str)s'
        func_args_str = base64.b64decode(encoded_func_args)
        self.args = loads(func_args_str, **unpickler_kw) or tuple()

        encoded_func_kwargs = '%(func_kwargs_str)s'
        func_kwargs_str = base64.b64decode(encoded_func_kwargs)
        self.kwargs = loads(func_kwargs_str, **unpickler_kw) or dict()

        self.from_types = '%(raw_from_type)s'.split(',')
        self.to_type = '%(to_type)s'

    def _handle_input(self, args):
        from datetime import datetime
        from decimal import Decimal

        res = []
        for t, arg in zip(self.from_types, args):
            if PY2 and t == 'datetime' and arg is not None and not isinstance(arg, datetime):
                res.append(datetime.fromtimestamp(arg / 1000.0))
            elif t == 'decimal' and arg is not None and isinstance(arg, str):
                res.append(Decimal(arg))
            else:
                res.append(arg)
        return res

    def _to_milliseconds(self, dt):
        return int((time.mktime(dt.timetuple()) + dt.microsecond/1000000.0) * 1000)

    def _handle_output(self, arg):
        from datetime import datetime
        from decimal import Decimal

        t = self.to_type

        if PY2 and t == 'datetime' and isinstance(arg, datetime):
            if isinstance(arg, np_generic):
                arg = arg.item()
            return self._to_milliseconds(arg)
        elif t == 'string' and isinstance(arg, Decimal):
            return str(arg)
        else:
            if isinstance(arg, np_generic):
                arg = arg.item()
            return arg

    def evaluate(self, %(input_args)s):
        args = %(input_args)s,
        args = self._handle_input(args)
        if not self.names:
            args = tuple(args) + tuple(self.args)
            res = self.f(*args, **self.kwargs)
            return self._handle_output(res)
        else:
            res = self.f(self.named_args(*args), *self.args, **self.kwargs)
            return self._handle_output(res)
'''


UDTF_TMPL = '''
import functools
from odps.udf import BaseUDTF

if PY2:
    string_type = unicode
    byte_type = str
else:
    string_type = str
    byte_type = (bytes, bytearray)

@annotate('%(from_type)s->%(to_type)s')
class %(func_cls_name)s(BaseUDTF):
    def __init__(self):
        unpickler_kw = dict(impl=%(implementation)s, dump_code=%(dump_code)s)
        rs = loads(base64.b64decode('%(resources)s'), **unpickler_kw)
        resources = []
        for t, n, fields in rs:
            if t == 'file':
                resources.append(get_cache_file(str(n)))
            elif t == 'archive':
                resources.append(get_cache_archive(str(n)))
            else:
                tb = get_cache_table(str(n))
                if fields:
                    tb = gen_resource_data(fields, tb)
                resources.append(tb)

        libraries = (l for l in '%(libraries)s'.split(',') if len(l) > 0)
        files = []
        for lib in libraries:
            if lib.startswith('a:'):
                lib = lib[2:]
                f = get_cache_archive_data(lib)
            else:
                f = get_cache_file(lib)
            files.append(read_lib(lib, f))
        sys.meta_path.append(CompressImporter(*files, supersede=%(supersede_libraries)r))

        load_np_generic()

        encoded = '%(func_str)s'
        f_str = base64.b64decode(encoded)
        self.f = loads(f_str, **unpickler_kw)

        if inspect.isclass(self.f):
            if not resources:
                self.f = self.f()
            else:
                self.f = self.f(resources)
            self.is_f_generator = inspect.isgeneratorfunction(self.f.__call__)
            self.close_f = getattr(self.f, 'close', None)
            self.is_close_f_generator = inspect.isgeneratorfunction(self.close_f)
        else:
            if resources:
                self.f = self.f(resources)

            if isinstance(self.f, functools.partial):
                self.is_f_generator = inspect.isgeneratorfunction(self.f.func)
            else:
                self.is_f_generator = inspect.isgeneratorfunction(self.f)
            self.close_f = None
            self.is_close_f_generator = False

        encoded_func_args = '%(func_args_str)s'
        func_args_str = base64.b64decode(encoded_func_args)
        self.args = loads(func_args_str, **unpickler_kw) or tuple()

        encoded_func_kwargs = '%(func_kwargs_str)s'
        func_kwargs_str = base64.b64decode(encoded_func_kwargs)
        self.kwargs = loads(func_kwargs_str, **unpickler_kw) or dict()

        self.names = tuple(it for it in '%(names_str)s'.split(',') if it)
        if self.names:
            self.name_args = xnamedtuple('NamedArgs', self.names)

        self.from_types = '%(raw_from_type)s'.split(',')
        self.to_types = '%(to_type)s'.split(',')

    def _handle_input(self, args):
        from datetime import datetime
        from decimal import Decimal

        res = []
        for t, arg in zip(self.from_types, args):
            if PY2 and t == 'datetime' and arg is not None and not isinstance(arg, datetime):
                res.append(datetime.fromtimestamp(arg / 1000.0))
            elif t == 'decimal' and arg is not None and not isinstance(arg, Decimal):
                res.append(Decimal(arg))
            else:
                res.append(arg)
        return res

    def _to_milliseconds(self, dt):
        return int((time.mktime(dt.timetuple()) + dt.microsecond/1000000.0) * 1000)

    def _handle_output(self, args):
        from datetime import datetime
        from decimal import Decimal

        if len(self.to_types) != len(args):
            raise ValueError('Function output size should be ' + str(len(self.to_types))
                + ', got ' + str(args))

        res = []
        for t, arg in zip(self.to_types, args):
            if PY2 and t == 'datetime' and isinstance(arg, datetime):
                if isinstance(arg, np_generic):
                    arg = arg.item()
                res.append(self._to_milliseconds(arg))
            elif t == 'string' and isinstance(arg, Decimal):
                res.append(str(arg))
            elif PY2 and t == 'string' and isinstance(arg, string_type):
                res.append(arg.encode('utf-8'))
            elif PY3 and t == 'string' and isinstance(arg, byte_type):
                res.append(arg.decode('utf-8'))
            else:
                if isinstance(arg, np_generic):
                    arg = arg.item()
                res.append(arg)
        return res

    def process(self, %(input_args)s):
        args = %(input_args)s,
        args = self._handle_input(args)
        if not self.names:
            args = tuple(args) + tuple(self.args)
        else:
            args = (self.name_args(*args), ) + tuple(self.args)

        if self.is_f_generator:
            for r in self.f(*args, **self.kwargs):
                if not isinstance(r, (list, tuple)):
                    r = (r, )
                self.forward(*self._handle_output(r))
        else:
            res = self.f(*args, **self.kwargs)
            if res:
                if not isinstance(res, (list, tuple)):
                    res = (res, )
                self.forward(*self._handle_output(res))

    def close(self):
        if not self.close_f:
            return

        if self.is_close_f_generator:
            for r in self.close_f(*self.args, **self.kwargs):
                if not isinstance(r, (list, tuple)):
                    r = (r, )
                self.forward(*self._handle_output(r))
        else:
            res = self.close_f(*self.args, **self.kwargs)
            if res:
                if not isinstance(res, (list, tuple)):
                    res = (res, )
                self.forward(*self._handle_output(res))
'''

UDAF_TMPL = '''
from odps.udf import BaseUDAF

@annotate('%(from_type)s->%(to_type)s')
class %(func_cls_name)s(BaseUDAF):
    def __init__(self):
        unpickler_kw = dict(impl=%(implementation)s, dump_code=%(dump_code)s)
        rs = loads(base64.b64decode('%(resources)s'), **unpickler_kw)
        resources = []
        for t, n, fields in rs:
            if t == 'file':
                resources.append(get_cache_file(str(n)))
            elif t == 'archive':
                resources.append(get_cache_archive(str(n)))
            else:
                tb = get_cache_table(str(n))
                if fields:
                    tb = gen_resource_data(fields, tb)
                resources.append(tb)

        libraries = (l for l in '%(libraries)s'.split(',') if len(l) > 0)
        files = []
        for lib in libraries:
            if lib.startswith('a:'):
                lib = lib[2:]
                f = get_cache_archive_data(lib)
            else:
                f = get_cache_file(lib)
            files.append(read_lib(lib, f))
        sys.meta_path.append(CompressImporter(*files, supersede=%(supersede_libraries)r))

        load_np_generic()

        encoded_func_args = '%(func_args_str)s'
        func_args_str = base64.b64decode(encoded_func_args)
        args = loads(func_args_str, **unpickler_kw) or tuple()

        encoded_func_kwargs = '%(func_kwargs_str)s'
        func_kwargs_str = base64.b64decode(encoded_func_kwargs)
        kwargs = loads(func_kwargs_str, **unpickler_kw) or dict()

        encoded = '%(func_str)s'
        f_str = base64.b64decode(encoded)
        agg = loads(f_str, **unpickler_kw)
        if resources:
            if not args and not kwargs:
                self.f = agg(resources)
            else:
                kwargs['resources'] = resources
                self.f = agg(*args, **kwargs)
        else:
            self.f = agg(*args, **kwargs)

        self.from_types = '%(raw_from_type)s'.split(',')
        self.to_type = '%(to_type)s'

    def _handle_input(self, args):
        from datetime import datetime
        from decimal import Decimal

        res = []
        for t, arg in zip(self.from_types, args):
            if PY2 and t == 'datetime' and arg is not None and not isinstance(arg, datetime):
                res.append(datetime.fromtimestamp(arg / 1000.0))
            elif t == 'decimal' and arg is not None and not isinstance(arg, Decimal):
                res.append(Decimal(arg))
            else:
                res.append(arg)
        return res

    def _to_milliseconds(self, dt):
        return int((time.mktime(dt.timetuple()) + dt.microsecond/1000000.0) * 1000)

    def _handle_output(self, arg):
        from datetime import datetime
        from decimal import Decimal

        t = self.to_type

        if PY2 and t == 'datetime' and isinstance(arg, datetime):
            if isinstance(arg, np_generic):
                arg = arg.item()
            return self._to_milliseconds(arg)
        elif t == 'string' and isinstance(arg, Decimal):
            return str(arg)
        else:
            if isinstance(arg, np_generic):
                arg = arg.item()
            return arg

    def new_buffer(self):
        return self.f.buffer()

    def iterate(self, buffer, %(input_args)s):
        args = %(input_args)s,
        args = self._handle_input(args)
        self.f(buffer, *args)

    def merge(self, buffer, pbuffer):
        self.f.merge(buffer, pbuffer)

    def terminate(self, buffer):
        res = self.f.getvalue(buffer)
        return self._handle_output(res)
'''


def _gen_map_udf(node, func_cls_name, libraries, func, resources,
                 func_to_udfs, func_to_resources, func_params):
    names_str = ''
    if isinstance(node, MappedExpr) and node._multiple and \
            all(f.name is not None for f in node.inputs):
        names_str = ','.join(f.name for f in node.inputs)
    from_type = ','.join(df_type_to_odps_type(t).name.replace('`', '') for t in node.input_types)
    to_type = df_type_to_odps_type(node.dtype).name.replace('`', '')
    raw_from_type = ','.join(
        df_type_to_odps_type(t).name.replace('`', '') for t in node.raw_input_types
    )
    func_args_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_args, dump_code=options.df.dump_udf))
    )
    func_kwargs_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_kwargs, dump_code=options.df.dump_udf))
    )

    if inspect.isfunction(func) and not func.__closure__:
        func_sig = id(func.__code__)
    else:
        func_sig = func

    key = (from_type, to_type, func_sig, tuple(resources), names_str, func_args_str, func_kwargs_str)
    if key in func_params:
        node.func = func_params[key]
        return
    else:
        if func in func_to_udfs:
            func = make_copy(func)
            node.func = func

        func_params[key] = func

    func_to_udfs[func] = UDF_TMPL_HEADER + UDF_TMPL % {
        'raw_from_type': raw_from_type,
        'from_type': from_type,
        'to_type': to_type,
        'func_cls_name': func_cls_name,
        'func_str': to_str(base64.b64encode(cloudpickle.dumps(func, dump_code=options.df.dump_udf))),
        'func_args_str': func_args_str,
        'func_kwargs_str': func_kwargs_str,
        'names_str': names_str,
        'resources': to_str(
            base64.b64encode(cloudpickle.dumps([r[:3] for r in resources], dump_code=options.df.dump_udf))),
        'implementation': CLIENT_IMPL,
        'dump_code': options.df.dump_udf,
        'input_args': ', '.join('arg{0}'.format(i) for i in range(len(node.input_types))),
        'libraries': ','.join(libraries if libraries is not None else []),
        'supersede_libraries': options.df.supersede_libraries,
    }
    if resources:
        func_to_resources[func] = resources


def _gen_apply_udf(node, func_cls_name, libraries, func, resources,
                   func_to_udfs, func_to_resources, func_params):
    names_str = ','.join(f.name for f in node.fields)
    from_type = ','.join(df_type_to_odps_type(t).name.replace('`', '') for t in node.input_types)
    raw_from_type = ','.join(
        df_type_to_odps_type(t).name.replace('`', '') for t in node.raw_input_types
    )
    to_type = ','.join(df_type_to_odps_type(t).name.replace('`', '') for t in node.schema.types)
    func_args_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_args, dump_code=options.df.dump_udf)))
    func_kwargs_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_kwargs, dump_code=options.df.dump_udf)))

    key = (from_type, to_type, func, tuple(resources), names_str, func_args_str, func_kwargs_str)
    if key in func_params:
        return
    else:
        if func in func_to_udfs:
            func = make_copy(func)
            node.func = func

        func_params[key] = func

    func_to_udfs[func] = UDF_TMPL_HEADER + UDTF_TMPL % {
        'raw_from_type': raw_from_type,
        'from_type': from_type,
        'to_type': to_type,
        'func_cls_name': func_cls_name,
        'func_str': to_str(base64.b64encode(cloudpickle.dumps(func, dump_code=options.df.dump_udf))),
        'func_args_str': func_args_str,
        'func_kwargs_str': func_kwargs_str,
        'close_func_str': to_str(
            base64.b64encode(cloudpickle.dumps(getattr(node, '_close_func', None), dump_code=options.df.dump_udf))),
        'names_str': names_str,
        'resources': to_str(base64.b64encode(cloudpickle.dumps([r[:3] for r in resources]))),
        'implementation': CLIENT_IMPL,
        'dump_code': options.df.dump_udf,
        'input_args': ', '.join('arg{0}'.format(i) for i in range(len(node.input_types))),
        'libraries': ','.join(libraries if libraries is not None else []),
        'supersede_libraries': options.df.supersede_libraries,
    }
    if resources:
        func_to_resources[func] = resources


def _gen_agg_udf(node, func_cls_name, libraries, func, resources,
                 func_to_udfs, func_to_resources, func_params):
    from_type = ','.join(df_type_to_odps_type(t).name.replace('`', '') for t in node.input_types)
    raw_from_type = ','.join(
        df_type_to_odps_type(t).name.replace('`', '') for t in node.raw_input_types
    )
    to_type = df_type_to_odps_type(node.dtype).name.replace('`', '')
    func_args_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_args, dump_code=options.df.dump_udf)))
    func_kwargs_str = to_str(
        base64.b64encode(cloudpickle.dumps(node._func_kwargs, dump_code=options.df.dump_udf)))

    key = (from_type, to_type, func, tuple(resources), func_args_str, func_kwargs_str)
    if key in func_params:
        return
    else:
        if func in func_to_udfs:
            func = make_copy(func)
            node.func = func

        func_params[key] = func

    func_to_udfs[func] = UDF_TMPL_HEADER + UDAF_TMPL % {
        'raw_from_type': raw_from_type,
        'from_type': from_type,
        'to_type': to_type,
        'func_cls_name': func_cls_name,
        'func_str': to_str(base64.b64encode(cloudpickle.dumps(func, dump_code=options.df.dump_udf))),
        'func_args_str': func_args_str,
        'func_kwargs_str': func_kwargs_str,
        'resources': to_str(
            base64.b64encode(cloudpickle.dumps([r[:3] for r in resources], dump_code=options.df.dump_udf))),
        'implementation': CLIENT_IMPL,
        'dump_code': options.df.dump_udf,
        'input_args': ', '.join('arg{0}'.format(i) for i in range(len(node.input_types))),
        'libraries': ','.join(libraries if libraries is not None else []),
        'supersede_libraries': options.df.supersede_libraries,
    }
    if resources:
        func_to_resources[func] = resources


def gen_udf(expr, func_cls_name=None, libraries=None):
    func_to_udfs = OrderedDict()
    func_to_resources = OrderedDict()
    func_params = dict()
    if libraries is not None:
        def _get_library_name(res):
            if isinstance(res, six.string_types):
                return res
            elif isinstance(res, ArchiveResource):
                return 'a:' + res.name
            else:
                return res.name

        libraries = [_get_library_name(lib) for lib in libraries]

    for node in expr.traverse(unique=True):
        func = getattr(node, 'func', None)
        if func is None:
            continue
        if isinstance(func, six.string_types):
            continue

        resources = []
        collection_idx = 0
        if hasattr(node, '_resources') and node._resources:
            for res in node._resources:
                if isinstance(res, ArchiveResource):
                    tp = 'archive'
                    name = res.name
                    fields = None
                    create = False
                    table_name = None
                elif isinstance(res, FileResource):
                    tp = 'file'
                    name = res.name
                    fields = None
                    create = False
                    table_name = None
                elif isinstance(res, TableResource):
                    tp = 'table'
                    name = res.name
                    fields = tuple(col.name for col in res.get_source_table().table_schema.simple_columns)
                    create = False
                    table_name = None
                else:
                    res = node._collection_resources[collection_idx]
                    collection_idx += 1

                    tp = 'table'
                    name = 'tmp_pyodps_resource_%s' % (uuid.uuid4())
                    fields = tuple(res.schema.names)
                    create = True
                    table_name = get_executed_collection_project_table_name(res)
                resources.append((tp, name, fields, create, table_name))

        if isinstance(node, MappedExpr):
            _gen_map_udf(node, func_cls_name, libraries, func, resources,
                         func_to_udfs, func_to_resources, func_params)
        elif isinstance(node, RowAppliedCollectionExpr):
            _gen_apply_udf(node, func_cls_name, libraries, func, resources,
                           func_to_udfs, func_to_resources, func_params)
        elif isinstance(node, (Aggregation, GroupedAggregation)):
            _gen_agg_udf(node, func_cls_name, libraries, func, resources,
                         func_to_udfs, func_to_resources, func_params)

    return func_to_udfs, func_to_resources



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import weakref
from collections import namedtuple

from ..utils import import_class_member
from ...compat import six
from ...utils import hashable
from ...df import DataFrame
from ...df.expr.collections import CollectionExpr, Expr, FilterPartitionCollectionExpr
from ...df.expr.dynamic import DynamicMixin, DynamicCollectionExpr
from ...utils import get_id

expr_output_registry = dict()
shared_props_registry = dict()


class AlgoExprProxy(object):
    def __init__(self, expr):
        def _finalizer(_):
            if self._exec_id not in expr_output_registry:
                return
            if self._register_name in expr_output_registry[self._exec_id]:
                del expr_output_registry[self._exec_id][self._register_name]
            if len(expr_output_registry[self._exec_id]) == 0:
                del expr_output_registry[self._exec_id]
                if self._exec_id in shared_props_registry:
                    del shared_props_registry[self._exec_id]

        self._ref = weakref.ref(expr, _finalizer)
        self._exec_id = expr._exec_id
        self._register_name = expr.register_name

    def __call__(self):
        return self._ref()

    def __getattr__(self, item):
        return getattr(self._ref(), item)


class AlgoExprMixin(Expr):
    _mixin_slots = '_exec_id', '_params', '_engine_kw', '_output_name', '_persist_kw', '_cases', \
                   '_table_callback', '_is_extra'
    _add_args_slots = False
    _algo = ''

    @staticmethod
    def cache_input(expr):
        if not isinstance(expr, Expr):
            return expr
        if isinstance(expr, FilterPartitionCollectionExpr) and isinstance(expr.input, DataFrame):
            return expr
        if isinstance(expr, AlgoExprMixin) and expr.executed:
            return expr
        return expr.cache(mem=False)

    def _register_name(self):
        return self._output_name

    register_name = property(fget=_register_name)

    def register_expr(self):
        if self._exec_id not in expr_output_registry:
            expr_output_registry[self._exec_id] = dict()
        expr_output_registry[self._exec_id][self.register_name] = AlgoExprProxy(self)
        shared_props_registry[self._exec_id] = dict()
        shared_props_registry[self._exec_id]['executed'] = False

    def uncache(self):
        pass

    def _get_executed(self):
        if self._exec_id not in shared_props_registry:
            return False
        return shared_props_registry[self._exec_id].get('executed', False)

    def _set_executed(self, value):
        if self._exec_id not in shared_props_registry:
            shared_props_registry[self._exec_id] = dict()
        shared_props_registry[self._exec_id]['executed'] = value
        if value and 'exec_lock' in shared_props_registry[self._exec_id]:
            exec_lock = shared_props_registry[self._exec_id]['exec_lock']
            exec_lock.release()

    executed = property(fget=lambda self: self._get_executed(),
                        fset=lambda self, value: self._set_executed(value))

    def wait_execution(self):
        if self._exec_id not in shared_props_registry:
            shared_props_registry[self._exec_id] = dict()
        if 'exec_lock' not in shared_props_registry[self._exec_id]:
            exec_lock = threading.Lock()
            shared_props_registry[self._exec_id]['exec_lock'] = exec_lock
        else:
            exec_lock = shared_props_registry[self._exec_id]['exec_lock']
        if not self.executed:
            exec_lock.acquire()

    def convert_params(self, src_expr=None):
        params = dict((k, v) for k, v in six.iteritems(self._params) if k in self._exported)

        for name, exporter in six.iteritems(self._exporters):
            if src_expr is not None:
                params[name] = exporter(src_expr)
            if not params[name]:
                params[name] = exporter(self)

        for k, v in list(six.iteritems(params)):
            if v is None:
                params.pop(k)
            if isinstance(v, (list, tuple, set)) and len(v) == 0:
                params.pop(k)
        return params

    @property
    def persist_kw(self):
        if self._exec_id not in expr_output_registry:
            raise AttributeError
        if self.register_name not in expr_output_registry[self._exec_id]:
            raise AttributeError
        return expr_output_registry[self._exec_id][self.register_name]._persist_kw

    @persist_kw.setter
    def persist_kw(self, value):
        if self._exec_id not in expr_output_registry:
            raise AttributeError
        if self.register_name not in expr_output_registry[self._exec_id]:
            raise AttributeError
        expr_output_registry[self._exec_id][self.register_name]._persist_kw = value

    @property
    def shared_kw(self):
        if self._exec_id not in shared_props_registry:
            raise AttributeError
        return shared_props_registry[self._exec_id].get('shared_kw', dict())

    @shared_kw.setter
    def shared_kw(self, value):
        if self._exec_id not in shared_props_registry:
            raise AttributeError
        shared_props_registry[self._exec_id]['shared_kw'] = value

    def outputs(self):
        if getattr(self, '_exec_id', None) is None:
            return dict()
        out_dict = expr_output_registry.get(self._exec_id, dict())
        out_dict = dict((k, v()) for k, v in six.iteritems(out_dict))
        return dict((k, v) for k, v in six.iteritems(out_dict) if v is not None and not v.is_extra_expr)

    @property
    def is_extra_expr(self):
        return getattr(self, '_is_extra', False)


class AlgoCollectionExpr(AlgoExprMixin, CollectionExpr):
    _suffix = 'CollectionExpr'
    node_name = 'Algo'

    def _init(self, *args, **kwargs):
        register_expr = kwargs.pop('register_expr', False)

        p_args = [a.cache() if isinstance(a, Expr) else a for a in args]
        p_kw = dict((k, self.cache_input(v)) for k, v in six.iteritems(kwargs))

        absent_args = (a for a in self._args[len(args):] if a not in p_kw)
        for a in absent_args:
            p_kw[a] = None

        super(AlgoCollectionExpr, self)._init(*p_args, **p_kw)
        self.cache(False)

        if register_expr:
            self.register_expr()

    def accept(self, visitor):
        visitor.visit_algo(self)


class DynamicAlgoCollectionExpr(DynamicMixin, AlgoCollectionExpr):
    def __init__(self, *args, **kwargs):
        DynamicMixin.__init__(*args, **kwargs)
        AlgoCollectionExpr.__init__(*args, **kwargs)

    _project = DynamicCollectionExpr._project


metrics_tables = dict()
metrics_executed = dict()


class MetricsResultExpr(AlgoExprMixin, Expr):
    __slots__ = '_metrics_hash', '_result_callback'
    _suffix = 'ResultExpr'
    _non_table = True
    node_name = 'Metrics'

    def _init(self, *args, **kwargs):
        kwargs.pop('register_expr', False)

        p_args = [a.cache() if isinstance(a, Expr) else a for a in args]
        p_kw = dict((k, self.cache_input(v)) for k, v in six.iteritems(kwargs))

        absent_args = (a for a in self._args[len(args):] if a not in p_kw)
        for a in absent_args:
            p_kw[a] = None

        super(MetricsResultExpr, self)._init(*p_args, **p_kw)

        if getattr(self, '_metrics_hash', None) is None:
            param_hash = hash(frozenset(six.iteritems(hashable(self._params))))
            port_hash = hash(frozenset(self.get_input_hash(pt) for pt in self.input_ports))
            self._metrics_hash = hash((self.node_name, param_hash, port_hash))

    def get_input_hash(self, input_port):
        from .exporters import get_ml_input
        input_expr = get_ml_input(self, input_port.name)
        pid = get_id(input_expr) if input_expr else None
        return hash((input_port.name, pid))

    @property
    def tables_tuple(self):
        cls = type(self)
        if not hasattr(cls, '_tables_tuple'):
            cls._tables_tuple = namedtuple('_tables_tuple', [pt.name for pt in cls.output_ports])
        return cls._tables_tuple

    @property
    def calculator(self):
        cls = type(self)
        if not hasattr(cls, '_calculator'):
            cls._calculator = import_class_member(self.algo_meta['calculator'])
        return self._calculator

    def accept(self, visitor):
        visitor.visit_algo(self)

    def _get_executed(self):
        return metrics_executed.get(self._metrics_hash, False)

    def _set_executed(self, value):
        metrics_executed[self._metrics_hash] = value

    @property
    def tables(self):
        return metrics_tables.get(self._metrics_hash)

    @tables.setter
    def tables(self, value):
        if isinstance(value, dict):
            value = self.tables_tuple(**value)
        metrics_tables[self._metrics_hash] = value


def make_extra_expr(expr, name, schema, **kw):
    new_expr = expr.copy(clear_keys=['_id'], register_expr=True, _is_extra=True, _schema=schema,
                         _output_name=name, **kw)
    new_expr.cache(False)
    return new_expr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from copy import deepcopy

from .compat import Queue, six, Iterable


class DAGValidationError(Exception):
    pass


class DAG(object):
    """Directed acyclic graph implementation."""
    _graph_dict_type = dict
    _dict_type = dict

    def __init__(self, reverse=False):
        self._graph = self._graph_dict_type()
        self._map = self._dict_type()
        if reverse:
            self._reversed_graph = self._graph_dict_type()
        else:
            self._reversed_graph = None

    def nodes(self):
        return [self._map[n] for n in self._graph]

    def contains_node(self, node):
        return id(node) in self._graph

    def add_node(self, node):
        self._graph[id(node)] = set()
        self._map[id(node)] = node
        if self._reversed_graph is not None:
            self._reversed_graph[id(node)] = set()

    def remove_node(self, node):
        if id(node) not in self._graph:
            raise KeyError('Node does not exist')

        self._graph.pop(id(node))
        self._map.pop(id(node))

        for edges in six.itervalues(self._graph):
            if id(node) in edges:
                edges.remove(id(node))

        if self._reversed_graph is not None:
            self._reversed_graph.pop(id(node))

            for edges in six.itervalues(self._reversed_graph):
                if id(node) in edges:
                    edges.remove(id(node))

    def contains_edge(self, predecessor_node, successor_node):
        if id(predecessor_node) not in self._graph or \
                id(successor_node) not in self._graph:
            return False

        return id(successor_node) in self._graph[id(predecessor_node)]

    def add_edge(self, predecessor_node, successor_node, validate=True):
        if id(predecessor_node) not in self._graph or \
                id(successor_node) not in self._graph:
            raise KeyError('Node does not exist')

        if validate:
            test_graph = deepcopy(self._graph)
            test_graph[id(predecessor_node)].add(id(successor_node))
            test_reversed_graph = None
            if self._reversed_graph is not None:
                test_reversed_graph = deepcopy(self._reversed_graph)
                test_reversed_graph[id(successor_node)].add(id(predecessor_node))
            valid, msg = self._validate(test_graph, test_reversed_graph)
        else:
            valid, msg = True, ''
        if valid:
            self._graph[id(predecessor_node)].add(id(successor_node))
            if self._reversed_graph is not None:
                self._reversed_graph[id(successor_node)].add(id(predecessor_node))
        else:
            raise DAGValidationError(msg)

    def remove_edge(self, predecessor_node, successor_node):
        if id(successor_node) not in self._graph.get(id(predecessor_node), []):
            raise KeyError('Edge does not exist in the graph')

        self._graph[id(predecessor_node)].remove(id(successor_node))
        if self._reversed_graph is not None:
            self._reversed_graph[id(successor_node)].remove(id(predecessor_node))

    def _indep_ids(self, graph=None, reversed_graph=None):
        graph = graph or self._graph
        reversed_graph = reversed_graph or self._reversed_graph

        if reversed_graph is not None:
            return [node for node, precessors in six.iteritems(reversed_graph)
                    if len(precessors) == 0]

        all_nodes = set(graph.keys())
        return list(all_nodes - set(itertools.chain(*graph.values())))

    def indep_nodes(self, graph=None, reversed_graph=None):
        return [self._map.get(i) for i in self._indep_ids(graph=graph,
                                                          reversed_graph=reversed_graph)]

    def _predecessor_ids(self, node_id, graph=None, reversed_graph=None):
        graph = graph or self._graph
        reversed_graph = reversed_graph or self._reversed_graph
        if reversed_graph is not None:
            return reversed_graph[node_id]
        return [nid for nid, deps in six.iteritems(graph) if node_id in deps]

    def predecessors(self, node):
        if id(node) not in self._graph:
            raise KeyError('Node does not exist: %s' % node)

        return [self._map.get(node_id) for node_id in self._predecessor_ids(id(node))]

    def successors(self, node):
        if id(node) not in self._graph:
            raise KeyError('Node does not exist: %r' % node)

        return [self._map.get(node_id) for node_id in self._graph[id(node)]]

    def _validate(self, graph=None, reversed_graph=None):
        graph = graph or self._graph
        reversed_graph = reversed_graph or self._reversed_graph
        if len(self.indep_nodes(graph, reversed_graph)) == 0:
            return False, 'No independent nodes detected'

        try:
            self.topological_sort(graph, reversed_graph)
        except ValueError:
            return False, 'Fail to topological sort'
        return True, 'Valid'

    def bfs(self, start_nodes, successor=None, cond=None):
        cond = cond or (lambda v: True)
        successor = successor or self.successors
        start_nodes = [start_nodes, ] if not isinstance(start_nodes, Iterable) else start_nodes
        start_nodes = [n for n in start_nodes if cond(n)]
        assert all(id(node) in self._graph for node in start_nodes)

        visited = set(id(node) for node in start_nodes)
        node_queue = Queue()
        [node_queue.put(node) for node in start_nodes]
        while not node_queue.empty():
            cur_node = node_queue.get()
            for up_node in (n for n in successor(cur_node) if cond(n)):
                if id(up_node) not in visited:
                    visited.add(id(up_node))
                    yield up_node
                    node_queue.put(up_node)

    def ancestors(self, start_nodes, cond=None):
        return list(self.bfs(start_nodes, self.predecessors, cond))

    def iter_ancestors(self, start_nodes, cond=None):
        for it in self.bfs(start_nodes, self.predecessors, cond=cond):
            yield it

    def descendants(self, start_nodes, cond=None):
        return list(self.bfs(start_nodes, cond=cond))

    def iter_descendants(self, start_nodes, cond=None):
        for it in self.bfs(start_nodes, cond=cond):
            yield it

    def topological_sort(self, graph=None, reversed_graph=None):
        graph = graph or self._graph
        graph = deepcopy(graph)
        reversed_graph = reversed_graph or self._reversed_graph
        reversed_graph = deepcopy(reversed_graph)

        node_ids = []

        indep_ids = self._indep_ids(graph, reversed_graph)
        while len(indep_ids) != 0:
            n = indep_ids.pop(0)
            node_ids.append(n)
            for dep_id in deepcopy(graph[n]):
                graph[n].remove(dep_id)
                if reversed_graph is not None:
                    reversed_graph[dep_id].remove(n)
                if len(self._predecessor_ids(dep_id, graph, reversed_graph)) == 0:
                    indep_ids.append(dep_id)

        if len(node_ids) != len(graph):
            raise ValueError('Graph is not acyclic')

        return [self._map.get(nid) for nid in node_ids]

    def reset_graph(self):
        self._graph = self._graph_dict_type()
        self._map = self._dict_type()
        if self._reversed_graph is not None:
            self._reversed_graph = self._graph_dict_type()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from .expressions import *
from .element import AnyOp, ElementWise
from . import utils
from .. import types


class BinOp(AnyOp):
    __slots__ = ()
    _args = '_lhs', '_rhs'

    @property
    def node_name(self):
        return self.__class__.__name__

    @property
    def name(self):
        if self._name:
            return self._name

        seq_hs = [hs for hs in self.args if isinstance(hs, SequenceExpr)]
        if len(seq_hs) == 1:
            return seq_hs[0].name

    def accept(self, visitor):
        visitor.visit_binary_op(self)


class UnaryOp(ElementWise):
    __slots__ = ()

    def accept(self, visitor):
        visitor.visit_unary_op(self)


class Arithmetic(BinOp):
    __slots__ = ()


class Comparison(BinOp):
    __slots__ = ()


class LogicalBinOp(BinOp):
    __slots__ = ()


class Negate(UnaryOp):
    __slots__ = ()


class Invert(UnaryOp):
    __slots__ = ()


class Abs(UnaryOp):
    __slots__ = ()


class Add(Arithmetic):
    __slots__ = ()


class Substract(Arithmetic):
    __slots__ = ()


class Multiply(Arithmetic):
    __slots__ = ()


class Divide(Arithmetic):
    __slots__ = ()


class FloorDivide(Arithmetic):
    __slots__ = ()


class Mod(Arithmetic):
    __slots__ = ()


class Power(Arithmetic):
    __slots__ = ()


class Greater(Comparison):
    __slots__ = ()


class Less(Comparison):
    __slots__ = ()


class Equal(Comparison):
    __slots__ = ()


class NotEqual(Comparison):
    __slots__ = ()


class GreaterEqual(Comparison):
    __slots__ = ()


class LessEqual(Comparison):
    __slots__ = ()


class Or(LogicalBinOp):
    __slots__ = ()


class And(LogicalBinOp):
    __slots__ = ()


def _get_type(other):
    if isinstance(other, SequenceExpr):
        other_type = other._data_type
    elif isinstance(other, Scalar):
        other_type = other._value_type
    else:
        other = Scalar(_value=other)
        other_type = other._value_type

    return other_type, other


def _arithmetic(expr, other, output_expr_cls, reverse=False, output_type=None):
    if isinstance(expr, (SequenceExpr, Scalar)):
        other_type, other = _get_type(other)
        is_sequence = isinstance(expr, SequenceExpr) or isinstance(other, SequenceExpr)
        if output_type is None:
            output_type = utils.highest_precedence_data_type(expr.dtype, other_type)

        if reverse:
            expr, other = other, expr
        if is_sequence:
            return output_expr_cls(_data_type=output_type, _lhs=expr, _rhs=other)
        else:
            return output_expr_cls(_value_type=output_type, _lhs=expr, _rhs=other)


def _reversed_arithmetic(expr, other, output_expr_cls, output_type=None):
    return _arithmetic(expr, other, output_expr_cls, reverse=True, output_type=output_type)


def _cmp(expr, other, output_expr_cls):
    if isinstance(expr, (SequenceExpr, Scalar)):
        other_type, other = _get_type(other)
        is_sequence = isinstance(expr, SequenceExpr) or isinstance(other, SequenceExpr)
        utils.highest_precedence_data_type(expr.dtype, other_type)
        # operand cast to data_type to compare
        output_type = types.boolean

        if is_sequence:
            return output_expr_cls(_data_type=output_type, _lhs=expr, _rhs=other)
        else:
            return output_expr_cls(_value_type=output_type, _lhs=expr, _rhs=other)


def _unary(expr, output_expr_cls):
    if isinstance(expr, (SequenceExpr, Scalar)):
        is_sequence = isinstance(expr, SequenceExpr)

        if is_sequence:
            return output_expr_cls(_data_type=expr.dtype, _input=expr)
        else:
            return output_expr_cls(_value_type=expr.dtype, _input=expr)


def _logic(expr, other, output_expr_cls):
    if isinstance(expr, (SequenceExpr, Scalar)):
        other_type, other = _get_type(other)
        is_sequence = isinstance(expr, SequenceExpr) or isinstance(other, SequenceExpr)

        if expr.dtype == types.boolean and other.dtype == types.boolean:
            output_type = types.boolean
            if is_sequence:
                return output_expr_cls(_data_type=output_type, _lhs=expr, _rhs=other)
            else:
                return output_expr_cls(_value_type=output_type, _lhs=expr, _rhs=other)

        raise TypeError('Logic operation needs boolean operand')


def _is_datetime(expr):
    if isinstance(expr, Expr):
        return expr.dtype == types.datetime
    else:
        return isinstance(expr, datetime)


def _add(expr, other):
    if _is_datetime(expr) and _is_datetime(other):
        raise ExpressionError('Cannot add two datetimes')
    return _arithmetic(expr, other, Add)


def _radd(expr, other):
    if _is_datetime(expr) and _is_datetime(other):
        raise ExpressionError('Cannot add two datetimes')
    return _reversed_arithmetic(expr, other, Add)


def _sub(expr, other):
    rtype = None
    if _is_datetime(expr) and _is_datetime(other):
        rtype = types.int64
    return _arithmetic(expr, other, Substract, output_type=rtype)


def _rsub(expr, other):
    rtype = None
    if _is_datetime(expr) and _is_datetime(other):
        rtype = types.int64
    return _reversed_arithmetic(expr, other, Substract, output_type=rtype)


def _eq(expr, other):
    return _cmp(expr, other, Equal)


def _ne(expr, other):
    return _cmp(expr, other, NotEqual)


def _gt(expr, other):
    return _cmp(expr, other, Greater)


def _lt(expr, other):
    return _cmp(expr, other, Less)


def _le(expr, other):
    return _cmp(expr, other, LessEqual)


def _ge(expr, other):
    return _cmp(expr, other, GreaterEqual)


def _mul(expr, other):
    return _arithmetic(expr, other, Multiply)


def _rmul(expr, other):
    return _reversed_arithmetic(expr, other, Multiply)


def _div(expr, other):
    if isinstance(expr.dtype, types.Integer) and isinstance(_get_type(other)[0], types.Integer):
        output_type = types.float64
    else:
        output_type = None
    return _arithmetic(expr, other, Divide, output_type=output_type)


def _rdiv(expr, other):
    if isinstance(expr.dtype, types.Integer) and isinstance(_get_type(other)[0], types.Integer):
        output_type = types.float64
    else:
        output_type = None
    return _reversed_arithmetic(expr, other, Divide, output_type=output_type)


def _mod(expr, other):
    return _arithmetic(expr, other, Mod)


def _rmod(expr, other):
    return _reversed_arithmetic(expr, other, Mod)


def _floordiv(expr, other):
    return _arithmetic(expr, other, FloorDivide)


def _rfloordiv(expr, other):
    return _reversed_arithmetic(expr, other, FloorDivide)


def _pow(expr, other):
    return _arithmetic(expr, other, Power)


def _rpow(expr, other):
    return _reversed_arithmetic(expr, other, Power)


def _or(expr, other):
    return _logic(expr, other, Or)


def _ror(expr, other):
    return _or(expr, other)


def _and(expr, other):
    return _logic(expr, other, And)


def _rand(expr, other):
    return _and(expr, other)


def _neg(expr):
    if isinstance(expr, Negate):
        return expr.input

    return _unary(expr, Negate)


def _invert(expr):
    if isinstance(expr, Invert):
        return expr.input

    return _unary(expr, Invert)


def _abs(expr):
    if isinstance(expr, Abs):
        return expr

    return _unary(expr, Abs)


_number_methods = dict(
    _add=_add,
    _radd=_radd,
    _sub=_sub,
    _rsub=_rsub,
    _mul=_mul,
    _rmul=_rmul,
    _div=_div,
    _rdiv=_rdiv,
    _floordiv=_floordiv,
    _rfloordiv=_rfloordiv,
    _mod=_mod,
    _rmod=_rmod,
    _pow=_pow,
    _rpow=_rpow,
    _neg=_neg,
    _abs=_abs,
    _eq=_eq,
    _ne=_ne,
    _gt=_gt,
    _lt=_lt,
    _le=_le,
    _ge=_ge,
)

_int_number_methods = dict(
    _invert=_invert
)

_string_methods = dict(
    _add=_add,
    _radd=_radd,
    _eq=_eq,
    _ne=_ne,
    _gt=_gt,
    _lt=_lt,
    _le=_le,
    _ge=_ge,
)

_boolean_methods = dict(
    _or=_or,
    _ror=_ror,
    _and=_and,
    _rand=_rand,
    _eq=_eq,
    _ne=_ne,
    _invert=_invert,
    _neg=_neg
)

_datetime_methods = dict(
    _add=_add,  # TODO, to check
    _radd=_radd,
    _sub=_sub,
    _rsub=_rsub,
    _eq=_eq,
    _ne=_ne,
    _gt=_gt,
    _lt=_lt,
    _le=_le,
    _ge=_ge
)

utils.add_method(StringSequenceExpr, _string_methods)
utils.add_method(StringScalar, _string_methods)

utils.add_method(BooleanSequenceExpr, _boolean_methods)
utils.add_method(BooleanScalar, _boolean_methods)

utils.add_method(DatetimeSequenceExpr, _datetime_methods)
utils.add_method(DatetimeScalar, _datetime_methods)

utils.add_method(DateSequenceExpr, _datetime_methods)
utils.add_method(DateScalar, _datetime_methods)

utils.add_method(TimestampSequenceExpr, _datetime_methods)
utils.add_method(TimestampScalar, _datetime_methods)

for number_sequence in number_sequences + number_scalars:
    utils.add_method(number_sequence, _number_methods)

for int_number_sequence in int_number_sequences + int_number_scalars:
    utils.add_method(int_number_sequence, _int_number_methods)
